from threading import Thread, Lock, BoundedSemaphore
import requests

class Driver(object):
    pass

class HttpDriver(object):
    max_threads = 4
    max_threads_semaphore = BoundedSemaphore(max_threads)

    @classmethod
    def retrieve(cls, uri):
        with cls.max_threads_semaphore:
            return requests.get(uri)

class FileProxy(object):
    def __init__(self, uri):
        self.uri = uri
        self.lock = Lock()
        self.data = None
        Thread(target=self.__retrieve).start()

    def stat(self):
        with self.lock:
            return self.data, self.data.headers

    def __retrieve(self):
        with self.lock:
            if self.uri.startswith('http://') or self.uri.startswith('https://'):
                self.data = HttpDriver.retrieve(self.uri)
            else:
                raise ValueError('Could not find driver for uri {0}.'.format(repr(self.uri)))

if __name__ == '__main__':
    urls = [
    ]

    files = [FileProxy(url) for url in urls]
    for file in files:
        print file.stat()




# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -*- coding: utf-8 -*-

from __future__ import absolute_import
import csv
from rdc.etl.io import STDIN
from rdc.etl.transform.map import Map
try:
    import cStringIO as StringIO
except:
    import StringIO


class CsvMap(Map):
    """
    Reads a CSV and yield the values, line-by-line.

    .. attribute:: delimiter

        The CSV delimiter.

    .. attribute:: quotechar

        The CSV quote character.

    .. attribute:: headers

        The list of column names, if the CSV does not contain it as its first line.

    .. attribute:: skip

        The amount of lines to skip before it actually yield output.

    """

    delimiter = ';'
    quotechar = '"'
    headers = None
    skip = 0

    def __init__(self, field=None, delimiter=None, quotechar=None, headers=None, skip=None):
        super(CsvMap, self).__init__(field=field)

        self.delimiter = str(delimiter or self.delimiter)
        self.quotechar = quotechar or self.quotechar
        self.headers = headers or self.headers
        self.skip = skip or self.skip

    @property
    def has_headers(self):
        return bool(self.headers)

    def transform(self, hash, channel=STDIN):
        s_in = StringIO.StringIO(hash.get(self.field))
        reader = csv.reader(s_in, delimiter=self.delimiter, quotechar=self.quotechar)
        headers = self.has_headers and self.headers or reader.next()
        field_count = len(headers)

        if self.skip and self.skip > 0:
            for i in range(0, self.skip):
                reader.next()

        for row in reader:
            if len(row) != field_count:
                raise ValueError('Got a line with %d fields, expecting %d.' % (len(row), field_count, ))

            yield hash.copy({self.field: row}).update(dict([(headers[i], row[i]) for i in range(0, field_count)]))




# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import copy

def default_comparator(a, b):
    if a == b: return 0
    if a < b: return -1
    if a > b: return 1

def default_merger(a, b):
    """Create a copy of first argument and update with second argument's value."""
    c = copy(a).update(b)
    return c

def get_lower(comparator, a, b):
    if comparator(a, b) > 0:
        return b
    return a

def get_higher(comparator, a, b):
    if comparator(a, b) > 0:
        return a
    return b

# TODO XXX isn't there some better algorithm for this ?
def insert_sorted(comparator, lst, key, value, start=0, end=None):
    # initial case, end is the end of list.
    if end is None:
        end = len(lst)

    if start == end:
        # 0 sized range
        lst.insert(start, (key, value, ))
        return start

    # compute length. end boundary excluded out of range
    length = end - start

    # find pivot
    pivot = start + length // 2

    # compare key with pivot's key
    side = comparator(key, lst[pivot][0])

    # equality ?
    if side == 0:
        lst.insert(pivot, (key, value, ))
        return pivot

    # left side ?
    if side == -1:
        # special case, only one value to consider (the pivot). Insert on the left.
        if length == 1:
            lst.insert(start, (key, value, ))
            return start

        # otherwise, consider the left hand interval (excluding pivot)
        return insert_sorted(comparator, lst, key, value, start, pivot)

    # right side ?
    if side == 1:
        # end of our sorted list ? append.
        if len(lst) <= pivot + 1:
            lst.append((key, value, ))
            return len(lst) - 1

        # special case, only one value to consider (the pivot). Insert on the right.
        if length == 1:
            lst.insert(start + 1, (key, value, ))
            return start + 1

        # otherwise, consider the right hand interval (excluding pivot)
        return insert_sorted(comparator, lst, key, value, pivot + 1, end)

    # wtf ?
    raise ValueError('Invalid comparator return value %r.' % (side, ))



# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from copy import copy

import sys
from rdc.etl.error import AbstractError
from rdc.etl.hash import Hash
from rdc.etl.io import STDIN, STDOUT, STDIN2, STDOUT2
from rdc.etl.transform import Transform
from rdc.etl.util import terminal as t

def shade(v):
    return t.black(t.bold(v))

def _repr(v):
    try:
        return unicode(v)
    except:
        try:
            return v.decode('utf-8')
        except:
            return repr(v)

class Log(Transform):
    """Identity transform that adds a console output side effect, to watch what is going through Queues at some point
    of an ETL process.

    """

    field_filter = None
    condition = None
    clean = False

    def __init__(self, field_filter=None, condition=None, clean=None):
        """Initializes the Log transform. Usually, no

        :param field_filter:
            An optional callable that will be used to filter which keys to display on transform.
        :param condition:
            An optional callable that will be used to determine whether or not this row should be logged (aka sent to
            stdout).
        """
        super(Log, self).__init__()

        self.field_filter = field_filter or self.field_filter
        self.condition = condition or self.condition
        self.clean = clean or self.clean

    def format(self, s):
        """Row formater."""

        # pretty format Hashes
        if isinstance(s, Hash):
            _s, s = s, []
            for k in _s.keys():
                s.append(u'  {k}{t.black}:{t.bold}{tp}{t.normal} {t.black}{t.bold}→{t.normal} {t.black}«{t.normal}{v}{t.black}»{t.normal}{t.clear_eol}'.format(k=_repr(k), v=_repr(_s[k]), t=t, tp=type(_s[k]).__name__))
        else:
            # unpack
            s = s.split('\n')

        # hack if one empty line
        if len(s) < 2 and not len(s[0].strip()):
            return ''

        # pack it back
        return '\n'.join(s)


    def writehr(self, label=None):
        if label:
            label = unicode(label)
            sys.stderr.write(t.black(u'·' * 4) + shade('{') + label + shade('}') + t.black(u'·' * (t.width - (6+len(label)) - 1)) + '\n')
        else:
            sys.stderr.write(t.black(u'·' * (t.width-1) + '\n'))


    def writeln(self, s):
        """Output method."""
        sys.stderr.write(self.format(s) + '\n')

    def initialize(self):
        self.lineno = 0

    def transform(self, hash, channel=STDIN):
        """Actual transformation."""
        self.lineno += 1
        if not self.condition or self.condition(hash):
            hash = hash.copy()
            hash = hash if not callable(self.field_filter) else hash.restrict(self.field_filter)
            if self.clean:
                hash = hash.restrict(lambda k: len(k) and k[0] != '_')
            self.writehr(self.lineno)
            self.writeln(hash)
            self.writehr()
            sys.stderr.write('\n')
        yield hash

class Limit(Transform):
    """Only pass the first `limit` input lines to default output.

    Args:
      limit (int): Number of line after which to stop passing input lines to the output.

    """

    limit = 0

    def __init__(self, limit = None):
        super(Limit, self).__init__()

        self.limit = limit or self.limit

    def initialize(self):
        super(Limit, self).initialize()
        self._current = 0

    def transform(self, hash, channel=STDIN):
        self._current += 1

        if self.limit and self._current <= self.limit:
            yield hash

class Stop(Transform):
    """Sinker transform that stops anything through the pipes.

    """

    def transform(self, hash, channel=STDIN):
        pass


class Override(Transform):
    """
    Simple transform that will overwrite some values with constant values provided in a Hash.

    """

    override_data = {}

    def __init__(self, override_data=None):
        super(Override, self).__init__()
        self.override_data = override_data or self.override_data

    def transform(self, hash, channel=STDIN):
        yield hash.update(self.override_data)

def clean(hash):
    for k in hash:
        if k.startswith('_'):
            del hash[k]
    return hash

class Clean(Transform):
    """
    Remove all fields with keys starting by _
    """

    def transform(self, hash, channel=STDIN):
        yield clean(hash)


class Validate(Transform):
    """An identity transform (input is passed to output unchanged). Aims for data validation."""

    CHANNEL_MAP = {
        STDIN: STDOUT,
        STDIN2: STDOUT2,
    }

    def __init__(self, validate=None):
        # Use the callable name if provided
        if validate and not self._name:
            self._name = validate.__name__

        # Parent constructor
        super(Validate, self).__init__()

        self.validate = validate or self.validate


    def validate(self, hash, channel=STDIN):
        raise AbstractError(self.validate)

    def transform(self, hash, channel=STDIN):
        # TODO instead of a copy, pass a "ReadOnly" decorated hash.
        self.validate(copy(hash), channel)
        yield hash, self.CHANNEL_MAP[STDIN]



# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
from rdc.etl.io import STDIN, STDIN2
from rdc.etl.transform import Transform
from rdc.etl.transform.flow import default_comparator, get_lower, insert_sorted, default_merger

class SortedJoin(Transform):
    INPUT_CHANNELS = (STDIN, STDIN2, )
    is_outer = False

    def __init__(self, key, merger=None, comparator=None, is_outer=None):
        super(SortedJoin, self).__init__()

        self.key = key
        self.merger = merger or default_merger
        self.comparator = comparator or default_comparator
        self.is_outer = is_outer or self.is_outer

    def initialize(self):
        self._minimum = None
        self._channel_value = dict(((c, None) for c in self.INPUT_CHANNELS))
        self._sorted = {
            STDIN: list(),
            STDIN2: list(),
        }

    def _get_minimum_value(self):
        return reduce(partial(get_lower, self.comparator), self._channel_value)

    def transform(self, hash, channel=STDIN):
        current_key = hash.get_values(self.key)
        insert_sorted(self.comparator, self._sorted[channel], current_key, hash)
        for data in self.consume():
            yield data

    def consume(self):
        while len(self._sorted[STDIN]) and len(self._sorted[STDIN2]):
            side = self.comparator(self._sorted[STDIN2][0][0], self._sorted[STDIN][0][0])

            if side == 0:
                # match, merge
                _key, _data = self._sorted[STDIN2].pop(0)
                pos = 0
                while (pos < (len(self._sorted[STDIN]) - 1)) and (self.comparator(self._sorted[STDIN][pos][0], _key) == 0):
                    self.merger(_data, self._sorted[STDIN][pos][1])
                    pos += 1
            elif side == -1:
                # no match
                if self.is_outer:
                    yield self._sorted[STDIN2].pop(0)[1]
                else:
                    self._sorted[STDIN2].pop(0)
            elif side == 1:
                # passed, yield possible
                yield self._sorted[STDIN].pop(0)[1]
            else:
                raise RuntimeError('invalid comparator return value')

    def finalize(self):
        for data in self.consume():
            yield data
        while len(self._sorted[STDIN]):
            yield self._sorted[STDIN].pop(0)[1]




# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

class AbstractError(NotImplementedError):
    """Abstract error is a convenient error to declare a method as "being left as an exercise for the reader"."""

    def __init__(self, method):
        super(AbstractError, self).__init__('Call to abstract method {class_name}.{method_name}(...): missing implementation.'.format(
            class_name=type(method.im_self).__name__,
            method_name=method.__name__,
        ))


class InactiveIOError(IOError):
    pass


class InactiveReadableError(InactiveIOError):
    pass


class InactiveWritableError(InactiveIOError):
    pass


class ValidationError(RuntimeError):
    def __init__(self, inst, message):
        super(ValidationError, self).__init__('Validation error in {class_name}: {message}'.format(
            class_name=type(inst).__name__,
            message=message,
        ))

class ProhibitedOperationError(RuntimeError):
    pass


# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from rdc.etl.util import slughifi


class UtilSlughifiTestCase(unittest.TestCase):
    def test_basic(self):
        self.assertEqual(slughifi('Test without unicode'), 'test-without-unicode')
        self.assertEqual(slughifi(u'Du français éèù!'), 'du-francais-eeu-')
        self.assertEqual(slughifi(u'Du français éèù!', strip=True), 'du-francais-eeu')


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import datetime
import psutil
from repoze.lru import lru_cache
from rdc.etl.status import BaseStatus
from rdc.etl.util import terminal as t


@lru_cache(1, timeout=1)
def memory_usage():
    process = psutil.Process(os.getpid())
    return process.get_memory_info()[0] / float(2 ** 20)


@lru_cache(64, timeout=1)
def execution_time(harness):
    return datetime.datetime.now() - harness._started_at


class ConsoleStatus(BaseStatus):
    """
    Outputs status information to the connected stdout. Can be a TTY, with or without support for colors/cursor
    movements, or a non tty (pipe, file, ...). The features are adapted to terminal capabilities.

    .. attribute:: prefix

        String prefix of output lines.

    """

    def __init__(self, prefix=''):
        self.prefix = prefix

    def _write(self, harness, debug, profile, rewind):
        threads = harness.get_threads()
        if profile:
            append = (
                (u'Memory', '{0:.2f} Mb'.format(memory_usage())),
                (u'Total time', '{0} s'.format(execution_time(harness))),
            )
        else:
            append = ()
        self.write(threads, prefix=self.prefix, append=append, debug=debug, profile=profile, rewind=rewind)

    def update(self, harness, debug, profile):
        if t.is_a_tty:
            self._write(harness, debug, profile, rewind=True)

    def finalize(self, harness, debug, profile):
        self._write(harness, debug, profile, rewind=False)

    @staticmethod
    def write(threads, prefix='', rewind=True, append=None, debug=False, profile=False):
        t_cnt = len(threads)

        for id, thread in threads:
            if thread.is_alive():
                _line = ''.join((t.black('({})'.format(id)), ' ', t.bold(t.white('+')), ' ', thread.name, ' ',
                                 thread.transform.get_unicode_stats(debug=debug, profile=profile), ' ', ))
            else:
                _line = t.black(''.join(
                    ('({})'.format(id), ' - ', thread.name, ' ', thread.transform.get_unicode_stats(debug=debug, profile=profile), ' ', )))
            print prefix + _line + t.clear_eol

        if append:
            # todo handle multiline
            print ' `->', ' '.join(u'{0}: {1}'.format(t.bold(t.white(k)), v) for k, v in append), t.clear_eol
            t_cnt += 1

        if rewind:
            print t.clear_eol
            print t.move_up * (t_cnt + 2)





# -*- coding: utf-8 -*-
#
# copyright 2012-2014 romain dorgueil
#
# licensed under the apache license, version 2.0 (the "license");
# you may not use this file except in compliance with the license.
# you may obtain a copy of the license at
#
#     http://www.apache.org/licenses/license-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the license is distributed on an "as is" basis,
# without warranties or conditions of any kind, either express or implied.
# see the license for the specific language governing permissions and
# limitations under the license.

from abc import ABCMeta, abstractmethod
from rdc.etl.error import AbstractError

class IHarness:
    """
    ETL harness interface.

    The harness is basically the executable stuff that will actually run a job.

    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def __call__(self):
        """Actual harness run."""
        raise AbstractError(self.__call__)

    @abstractmethod
    def add(self, transform):
        raise AbstractError(self.add)

    @abstractmethod
    def loop(self):
        raise AbstractError(self.loop)


class BaseHarness(IHarness):
    def initialize(self):
        pass

    def validate(self):
        pass

    def finalize(self):
        pass

    def __call__(self):
        """Implements IHarness.__call__()"""
        self.initialize()
        self.validate()
        _value = self.loop()
        self.finalize()
        return _value


# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rdc.etl.error import AbstractError
from rdc.etl.io import STDIN
from rdc.etl.transform import Transform


class Filter(Transform):
    """Filter out hashes from the stream depending on the :attr:`filter` callable return value, when called with the
    current hash as parameter.

    Can be used as a decorator on a filter callable.

    .. attribute:: filter

        A callable used to filter the hashes. If return value is True for a given hash, then the hash will be yield to
        output. Otherwise, it will be burnt.


    Example::

        >>> from rdc.etl.transform.filter import Filter
        >>> from rdc.etl.hash import Hash

        >>> @Filter
        ... def my_filter(hash, channel):
        ...     return hash['keepme'] == True

        >>> list(my_filter(
        ...         (('foo', 'bar'), ('keepme', True), ),
        ...         (('foo', 'baz'), ('keepme', False), ),
        ...     ))
        [H{'foo': 'bar', 'keepme': True}]

    """

    def __init__(self, filter=None):
        super(Filter, self).__init__()
        self.filter = filter or self.filter

    def filter(self, hash, channel=STDIN):
        raise AbstractError(self.filter)

    def transform(self, hash, channel=STDIN):
        if not self.filter or not callable(self.filter):
            raise RuntimeError('No callable provided to ' + self.__class__.__name__ + '.')

        if self.filter(hash, channel):
            yield hash





# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod
from rdc.etl.error import AbstractError


class IStatus:
    __metaclass__ = ABCMeta

    @abstractmethod
    def initialize(self, harness, debug, profile):
        """Initialize status."""
        raise AbstractError(self.initialize)

    @abstractmethod
    def update(self, harness, debug, profile):
        """update this status"""
        raise AbstractError(self.update)

    @abstractmethod
    def finalize(self, harness, debug, profile):
        """Finalize status."""
        raise AbstractError(self.finalize)


class BaseStatus(IStatus):
    def initialize(self, harness, debug, profile):
        pass

    def update(self, harness, debug, profile):
        raise AbstractError(self.update)

    def finalize(self, harness, debug, profile):
        pass





# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import OrderedDict

import unittest
from rdc.etl.extra.unittest import BaseTestCase
from rdc.etl.hash import Hash
from rdc.etl.transform.filter import Filter

INPUT_DATA = (
    OrderedDict((('foo', 'bar'), ('keepme', True), )),
    OrderedDict((('foo', 'baz'), ('keepme', False), )),
)

class TransformFilterTestCase(BaseTestCase):
    def test_base_class_decorator(self):
        @Filter
        def my_filter(hash, channel):
            return hash['keepme'] == True

        out = my_filter(*INPUT_DATA)

        self.assertStreamEqual(out, (INPUT_DATA[0], ))


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Maps are transforms that will yield rows depending on the value of one input field. In association with ``FileExtract``
for example, it can parse the file content format and yield rows that have an added knowledge.

By default, maps use the topic (`_`) field for input

"""

from rdc.etl import DEFAULT_FIELD
from rdc.etl.error import AbstractError
from rdc.etl.io import STDIN
from rdc.etl.transform import Transform


class Map(Transform):
    """Base class for mappers.

    .. attribute:: map

        Map logic callable. Takes the hash's field value and yields iterable data.

    .. attribute:: field

        The input field.


    Example::

        >>> from rdc.etl.transform.map import Map
        >>> from rdc.etl.transform.util import clean

        >>> @Map
        ... def my_map(s_in):
        ...     for l in s_in.split('\\n'):
        ...        yield {'f%d' % i: v for i, v in enumerate(l.split(':'))}

        >>> map(clean, my_map({'_': 'a:b:c\\nb:c:d\\nc:d:e'}))
        [H{'f0': 'a', 'f1': 'b', 'f2': 'c'}, H{'f0': 'b', 'f1': 'c', 'f2': 'd'}, H{'f0': 'c', 'f1': 'd', 'f2': 'e'}]

    """

    field = DEFAULT_FIELD

    def __init__(self, map=None, field=None):
        super(Map, self).__init__()

        self.map = map or self.map
        self.field = field or self.field

    def map(self, value):
        raise AbstractError(self.map)

    def transform(self, hash, channel=STDIN):
        for line in self.map(hash[self.field]):
            yield hash.copy(line)




# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from copy import copy

from sqlalchemy import MetaData, Table
from rdc.etl.error import ProhibitedOperationError
from rdc.etl.hash import Hash
from rdc.etl.io import STDIN, INSERT, UPDATE, SELECT, STDERR
from rdc.etl.transform import Transform
from rdc.etl.util import now, cached_property


class DatabaseLoad(Transform):
    """

    TODO doc this !!! test this !!!!

    """

    engine = None
    table_name = None
    fetch_columns = None
    insert_only_fields = ()
    discriminant = ('id', )
    created_at_field = 'created_at'
    updated_at_field = 'updated_at'
    allowed_operations = (INSERT, UPDATE, )

    def __init__(self, engine=None, table_name=None, fetch_columns=None, discriminant=None, created_at_field=None,
                 updated_at_field=None, insert_only_fields=None, allowed_operations=None):
        super(DatabaseLoad, self).__init__()

        self.engine = engine or self.engine
        self.table_name = table_name or self.table_name

        # xxx should take self.fetch_columns into account if provided
        self.fetch_columns = {}
        if isinstance(fetch_columns, list):
            self.add_fetch_column(*fetch_columns)
        elif isinstance(fetch_columns, dict):
            self.add_fetch_column(**fetch_columns)

        self.discriminant = discriminant or self.discriminant
        self.created_at_field = created_at_field or self.created_at_field
        self.updated_at_field = updated_at_field or self.updated_at_field
        self.insert_only_fields = insert_only_fields or self.insert_only_fields
        self.allowed_operations = allowed_operations or self.allowed_operations

        self._buffer = []
        self._connection = None
        self._max_buffer_size = 100
        self._last_duration = None
        self._last_commit_at = None
        self._query_count = 0

    @property
    def connection(self):
        if self._connection is None:
            self._connection = self.engine.connect()
        return self._connection

    def commit(self):
        with self.connection.begin():
            while len(self._buffer):
                hash = self._buffer.pop(0)
                try:
                    yield self.do_transform(copy(hash))
                except Exception as e:
                    yield Hash((
                        ('_input', hash, ),
                        ('_transform', self, ),
                        ('_error', e, ),
                    )), STDERR

    def close_connection(self):
        self._connection.close()
        self._connection = None

    def get_insert_columns_for(self, hash):
        """List of columns we can use for insert."""
        return self.columns

    def get_update_columns_for(self, hash, row):
        """List of columns we can use for update."""
        return [
            column for column in self.columns
            if not column in self.insert_only_fields
        ]

    def get_columns_for(self, hash, row=None):
        """Retrieve list of table column names for which we have a value in given hash.

        """
        if row:
            column_names = self.get_update_columns_for(hash, row)
        else:
            column_names = self.get_insert_columns_for(hash)

        return [key for key in hash if key in column_names]

    def find(self, dataset, connection=None):
        query = '''SELECT * FROM {table} WHERE {criteria} LIMIT 1'''.format(
            table=self.table_name,
            criteria=' AND '.join([key_atom + ' = %s' for key_atom in self.discriminant]),
        )
        rp = (connection or self.connection).execute(query, [dataset.get(key_atom) for key_atom in self.discriminant])

        # Increment stats
        self._input._special_stats[SELECT] += 1

        return rp.fetchone()

    def initialize(self):
        super(DatabaseLoad, self).initialize()

        self._input._special_stats[SELECT] = 0
        self._output._special_stats[INSERT] = 0
        self._output._special_stats[UPDATE] = 0


    def do_transform(self, hash):
        """Actual database load transformation logic, without the buffering / transaction logic.

        """

        # find line, if it exist
        row = self.find(hash)

        now = self.now
        column_names = self.table.columns.keys()
        # UpdatedAt field configured ? Let's set the value in source hash
        if self.updated_at_field in column_names:
            hash[self.updated_at_field] = now
        # Otherwise, make sure there is no such field
        else:
            if self.updated_at_field in hash:
                del hash[self.updated_at_field]

        # UPDATE
        if row:
            if not UPDATE in self.allowed_operations:
                raise ProhibitedOperationError('UPDATE operations are not allowed by this transformation.')

            _columns = self.get_columns_for(hash, row)

            query = '''UPDATE {table} SET {values} WHERE {criteria}'''.format(
                table=self.table_name,
                values=', '.join((
                    '{column} = %s'.format(column=_column) for _column in _columns
                    if not _column in self.discriminant
                )),
                criteria=' AND '.join((
                    '{key} = %s'.format(key=_key) for _key in self.discriminant
                ))
            )
            values = [hash[_column] for _column in _columns if not _column in self.discriminant] + \
                     [hash[_column] for _column in self.discriminant]

        # INSERT
        else:
            if not INSERT in self.allowed_operations:
                raise ProhibitedOperationError('INSERT operations are not allowed by this transformation.')

            if self.created_at_field in column_names:
                hash[self.created_at_field] = now
            else:
                if self.created_at_field in hash:
                    del hash[self.created_at_field]

            _columns = self.get_columns_for(hash)
            query = '''INSERT INTO {table} ({keys}) VALUES ({values})'''.format(
                table=self.table_name,
                keys=', '.join(_columns),
                values=', '.join(['%s'] * len(_columns))
            )
            values = [hash[key] for key in _columns]

        # Execute
        self.connection.execute(query, values)

        # Increment stats
        if row:
            self._output._special_stats[UPDATE] += 1
        else:
            self._output._special_stats[INSERT] += 1


        # If user required us to fetch some columns, let's query again to get their actual values.
        if self.fetch_columns and len(self.fetch_columns):
            if not row:
                row = self.find(hash)
            if not row:
                raise ValueError('Could not find matching row after load.')

            for alias, column in self.fetch_columns.iteritems():
                hash[alias] = row[column]

        return hash

    def transform(self, hash, channel=STDIN):
        """Transform method. Stores the input in a buffer, and only unstack buffer content if some limit has been
        exceeded.

        TODO for now buffer limit is hardcoded as 1000, but we may use a few criterias to add intelligence to this:
             time since last commit, duration of last commit, buffer length ...

        """
        self._buffer.append(hash)

        if len(self._buffer) >= self._max_buffer_size:
            for _out in self.commit():
                yield _out

    def finalize(self):
        """Transform's finalize method.

        Empties the remaining lines in buffer by loading them into database and close database connection.

        """

        super(DatabaseLoad, self).finalize()

        for _out in self.commit():
            yield _out

        self.close_connection()

    def add_fetch_column(self, *columns, **aliased_columns):
        self.fetch_columns.update(aliased_columns)
        for column in columns:
            self.fetch_columns[column] = column

    @cached_property
    def columns(self):
        return self.table.columns.keys()

    @cached_property
    def metadata(self):
        """SQLAlchemy metadata."""
        return MetaData()

    @cached_property
    def table(self):
        """SQLAlchemy table object, using metadata autoloading from database to avoid the need of column definitions."""
        return Table(self.table_name, self.metadata, autoload=True, autoload_with=self.engine)

    @property
    def now(self):
        """Current timestamp, used for created/updated at fields."""
        return now()




# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import OrderedDict

import unittest
from rdc.etl.extra.unittest import BaseTestCase
from rdc.etl.hash import Hash
from rdc.etl.transform.map import Map


class TransformMapTestCase(BaseTestCase):
    def test_base_class_decorator(self):
        @Map
        def my_map(s):
            for l in s.split('\n'):
                yield Hash((('f%d' % (i, ), v) for i, v in enumerate(l.split(':'))))

        self.assertStreamEqual(
            my_map((Hash({'_': 'a:b:c\nb:c:d\nc:d:e'}))), (
                OrderedDict((('f0', 'a'), ('f1', 'b'), ('f2', 'c'), ), ),
                OrderedDict((('f0', 'b'), ('f1', 'c'), ('f2', 'd'), ), ),
                OrderedDict((('f0', 'c'), ('f1', 'd'), ('f2', 'e'), ), ),
            ))


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from webapp2 import RequestHandler
from wsgiref.simple_server import make_server
from threading import Thread
import webapp2
from rdc.etl.status import BaseStatus

class HttpHandler(RequestHandler):
    @property
    def harness(self):
        return self.app.config.get('harness')

    def get(self):
        threads = self.harness._threads.items()

        self.response.write('''<!DOCTYPE html>
<!--[if IE 9]><html class="lt-ie10" lang="en" > <![endif]-->
<html class="no-js" lang="en" >

<head>
    <meta charset="utf-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>rdc.etl status</title>
    <link rel="stylesheet" href="//cdnjs.cloudflare.com/ajax/libs/foundation/5.0.2/css/normalize.min.css">
    <link rel="stylesheet" href="//cdnjs.cloudflare.com/ajax/libs/foundation/5.0.2/css/foundation.min.css">
    <script src="//cdnjs.cloudflare.com/ajax/libs/modernizr/2.7.1/modernizr.min.js"></script>
    <style>
        .alive {

        }
        .dead {
            color: grey;
        }
        .thread {
            padding: 0.25rem;
            margin-bottom: 0.25rem;
            background-color: white;
        }

    </style>
</head>

<body>
    <div class="row">
        <div class="right" style="margin-top: 20px">
            <a class="button tiny round" href="/">reload</a>
            <a class="button tiny round alert" href="/halt">halt</a>
        </div>
        <h1>rdc.etl - web status</h1>
    </div>
    <div class="row">
''')

        finished = False
        for id, thread in threads:
            self.response.write('''<div class="panel thread {alive}">
                (<b>{id}</b>) {name} {stats}
            </div>'''.format(
                id=id,
                alive= 'alive' if thread.is_alive() else 'dead',
                name=thread.name,
                stats=thread.transform.get_stats_as_string()
            ))

        self.response.write('''
    </div>
    <script src="//cdnjs.cloudflare.com/ajax/libs/zepto/1.1.1/zepto.min.js"></script>
    <script src="//cdnjs.cloudflare.com/ajax/libs/foundation/5.0.2/js/foundation.min.js"></script>
    <script>$(document).foundation();</script>
</body>
</html>''')

class HttpServerThread(Thread):
    daemon = True

    def __init__(self, harness, verbose=None):
        super(HttpServerThread, self).__init__(verbose=verbose)
        self.wsgi_app = webapp2.WSGIApplication([
            ('/', HttpHandler),
        ], debug=True, config={
            'harness': harness
        })

    def run(self):
        self._wsgi_server = make_server('0.0.0.0', 5050, app=self.wsgi_app)
        self._wsgi_server.serve_forever()

class HttpStatus(BaseStatus):
    def initialize(self, harness, debug, profile):
        self.server = HttpServerThread(harness)
        self.server.start()

    def update(self, harness, debug, profile):
        pass





# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rdc.etl.io import STDIN
from rdc.etl.transform.join import Join


class DatabaseJoin(Join):
    """
    Operates a simple cartesian product between a hash and a row collection coming from a dynamic database query.

    For now, you need to pass an sqlite engine to contructor, as the work on external services connections won't be in
    milestone 1.0.

    .. attribute:: query

        SQL query template used to join flux data with database relation.

    .. attribute:: dataset_keys_for_values

        List of hash keys used to populate %-values in sql. For example, if query == "SELECT name FROM user WHERE id = %s",
        this tuple can be ('id', ), meaning that for each input line, the "id" field value will be used to execute
        the statement. Each sql statement output row will be joined with the input row.

    """

    query = None
    dataset_keys_for_values = []

    def __init__(self, engine, query=None, dataset_keys_for_values=None, is_outer=False, default_outer_join_data=None):
        super(DatabaseJoin, self).__init__(is_outer=is_outer, default_outer_join_data=default_outer_join_data)

        # parameters
        self.engine = engine
        self.query = query or self.query
        self.dataset_keys_for_values = dataset_keys_for_values or self.dataset_keys_for_values

        # database connection
        self._connection = None

    def join(self, hash, channel=STDIN):
        """Get data to join with from database."""
        return self.connection.execute(self.query, [
            hash[key] for key in self.dataset_keys_for_values
        ])

    def finalize(self):
        """
        Finalize the transformation.

        DBAPI connection should be cleaned up.

        """
        super(DatabaseJoin, self).finalize()

        self._close_connection()

    @property
    def connection(self):
        """
        Database connection. If not done yet, connect().

        TODO: check what happens if for some reason the connection has been closed but attribute was not set to None.

        """
        if not self._connection:
            self._connection = self.engine.connect()
        return self._connection

    def _close_connection(self):
        """
        Closes database connection.

        """
        if self._connection:
            self._connection.close()
        self._connection = None






# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from copy import copy
from inspect import isgenerator
from rdc.etl.error import AbstractError
from rdc.etl.hash import Hash
from rdc.etl.transform.map import Map
from rdc.etl.util import etree


class XmlMap(Map):
    """
    Reads a XML and yield values for each root children.

    .. todo:: think how we want to make this flexible, xpath, etc ...
    .. warning:: This does not work, don't use (or fix before :p).

    Definitions:

        XML Item: In the context of an XmlMap, we define an XML Item as being either a children of the XML root if no
        xpath has been provided, or one item returned by the XPath provided.

    .. attribute:: map_item

        Will be called for each input XML Item, and should return a dictionary of values for this item.

    .. attribute:: field

        The input field (defined in parent).

    .. attribute:: xpath

        XPath used to select items before running them through item_map().

    """

    xpath = None

    def __init__(self, map_item=None, xpath=None, field=None):
        super(XmlMap, self).__init__()

        self.map_item = map_item or self.map_item
        self.xpath = xpath or self.xpath
        self.field = field or self.field

    def map(self, value):
        """Generator that yields mapped items.

        """
        items = etree.fromstring(value)

        if self.xpath:
            items = items.findall(self.xpath)

        for item in items:
            mapped = Hash(((self.field, item, ), ))
            value_for_item = self.map_item(item)
            if value_for_item:
                if isgenerator(value_for_item):
                    for _value in value_for_item:
                        if _value:
                            yield copy(mapped).update(_value)
                else:
                    try:
                        mapped.update(value_for_item)
                        yield mapped
                    except TypeError, e:
                        raise TypeError('{name}.map_item(...) must be iterable.'.format(
                            name=type(self).__name__
                        ))

    def map_item(self, item):
        """Convert one matched XML item to a dictionary.

        """
        raise AbstractError(self.map_item)





# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rdc.etl.error import AbstractError
from rdc.etl.hash import Hash
from rdc.etl.io import STDIN
from rdc.etl.transform import Transform


class Join(Transform):
    """
    Join some key => value pairs, that can depend on the source hash.

    This element can change the stream length, either positively (joining >1 item data) or negatively (joining <1 item data)

    .. automethod:: join

    Example::

        >>> from rdc.etl.transform.join import Join
        >>> from rdc.etl.transform.util import clean

        >>> @Join
        ... def my_join(hash, channel=STDIN):
        ...     return ({'a':1}, {'b':2}, )

        >>> map(clean, my_join({'foo': 'bar'}, {'foo': 'baz'}, ))
        [H{'foo': 'bar', 'a': 1}, H{'foo': 'bar', 'b': 2}, H{'foo': 'baz', 'a': 1}, H{'foo': 'baz', 'b': 2}]

    """

    is_outer = False

    """
    Return default join data when an outer join is requested but join data is empty. Not used in the default inner
    join case, because no row will be returned if current row did not generate join data.

    """
    default_outer_join_data = Hash()

    def __init__(self, join = None, is_outer=False, default_outer_join_data=None):
        super(Join, self).__init__()
        self.is_outer = is_outer or self.is_outer
        self.default_outer_join_data = default_outer_join_data or self.default_outer_join_data
        self.join = join or self.join

    def join(self, hash, channel=STDIN):
        """
        Abtract method that must be implemented in concrete subclasses, to return the data that should be joined with
        the given row.

        It should be iterable, or equivalent to False in a test.

        If the result is iterable and its length is superior to 0, the result of this transform will be a cartesian
        product between this method result and the original input row.

        If the result is false or iterable but 0-length, the result of this transform will depend on the join type,
        determined by the is_outer attribute.

        - If is_outer == True, the transform output will be a simple union between the input row and the result of
          self.get_default_outer_join_data()
        - If is_outer == False, this row will be sinked, and will not generate any output from this transform.

        Default join type is inner, to preserve backward compatibility.

        """
        raise AbstractError(self.join)

    def transform(self, hash, channel=STDIN):
        join_data = self.join(hash, channel)

        cnt = 0
        if join_data:
            for data in join_data:
                yield hash.copy(data)
                cnt += 1

        if not cnt and self.is_outer:
            yield hash.copy(self.default_outer_join_data)




# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rdc.etl.io import STDOUT, STDOUT2, STDERR, STDIN
from rdc.etl.transform import Transform


class Split(Transform):
    """Depending on the result of output_selector(), routes the input rows to various output channels."""

    OUTPUT_CHANNELS = (STDOUT, STDOUT2, STDERR, )

    def __init__(self, output_selector = None):
        super(Split, self).__init__()

        if output_selector:
            self.select_output = output_selector

    def output_selector(self, hash):
        return STDOUT

    def transform(self, hash, channel=STDIN):
        output_channel = self.select_output(hash)

        if not output_channel in self.OUTPUT_CHANNELS:
            raise IOError('Selected output channel %r does not exist.' % (output_channel, ))

        yield hash, output_channel


# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import copy
from rdc.etl import DEFAULT_FIELD
from rdc.etl.error import AbstractError
from rdc.etl.io import STDIN
from rdc.etl.transform import Transform


class SplitMap(Transform):
    """
    Split using a function.

    .. stability:: alpha

    """

    field = DEFAULT_FIELD
    _output_field = None

    def __init__(self, split=None, field=None, output_field=None):
        super(SplitMap, self).__init__()

        self.split = split or self.split
        self.field = field or self.field
        self._output_field = output_field or self._output_field

    @property
    def output_field(self):
        return self._output_field or self.field

    @output_field.setter
    def output_field(self, value):
        self._output_field = value

    def split(self, field):
        raise AbstractError(self.split)

    def transform(self, hash, channel=STDIN):
        values = self.split(hash[self.field])
        del hash[self.field]
        for value in values:
            hash = copy(hash)
            hash[self.output_field] = value
            yield hash




from paste.script.templates import Template, var

vars = [
    var('version', '0.1'),
    var('description', 'ETL Project'),
    var('long_description', 'This project was autogenerated by {0}.'.format(__package__)),
    var('keywords', 'Space-separated keywords/tags'),
    var('author', 'Author name'),
    var('author_email', 'Author email'),
    var('url', 'URL of homepage'),
    var('license_name', 'License name'),
    var('zip_safe', 'True/False: if the package can be distributed as a .zip file',
        default=False),
]

class ETLProjectTemplate(Template):
    _template_dir = 'templates/project'
    summary = 'ETL Project'
    vars = vars


# -*- coding: utf-8 -*-
#
# copyright 2012-2014 romain dorgueil
#
# licensed under the apache license, version 2.0 (the "license");
# you may not use this file except in compliance with the license.
# you may obtain a copy of the license at
#
#     http://www.apache.org/licenses/license-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the license is distributed on an "as is" basis,
# without warranties or conditions of any kind, either express or implied.
# see the license for the specific language governing permissions and
# limitations under the license.

from functools import wraps

def TransformBuilder(cls, *args, **kwargs):
    def wrapper(f):
        @wraps(f)
        def wrapped_constructor(*more_args, **more_kwargs):
            kwargs.update(more_kwargs)
            return cls(f, *(args+more_args), **kwargs)
        return wrapped_constructor
    return wrapper





# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from Queue import Empty
from rdc.etl.io import Input, InactiveWritableError, Begin, End, InactiveReadableError, InputMultiplexer


class InputTestCase(unittest.TestCase):
    def test_input(self):
        q = Input()

        # Before Begin, noone should be able to write in an Input queue.
        self.assertEqual(q.alive, False)
        self.assertRaises(InactiveWritableError, q.put, 'foo')

        # Begin
        q.put(Begin)
        self.assertEqual(q.alive, True)
        self.assertEqual(q._runlevel, 1)
        q.put('foo')

        # Second Begin
        q.put(Begin)
        self.assertEqual(q.alive, True)
        self.assertEqual(q._runlevel, 2)
        q.put('bar')
        q.put(End)

        # FIFO
        self.assertEqual(q.get(), 'foo')
        self.assertEqual(q.get(), 'bar')

        # self.assertEqual(q.alive, False) XXX queue don't know it's dead yet, but it is ...
        # Async get raises Empty (End is not returned)
        self.assertRaises(Empty, q.get, block=False)
        self.assertEqual(q.alive, True)

        # Before killing, let's slide some data in.
        q.put('baz')

        # now kill the queue, try to write
        q.put(End)
        self.assertRaises(InactiveWritableError, q.put, 'foo')

        # still can get remaining data
        self.assertEqual(q.get(), 'baz')
        self.assertRaises(InactiveReadableError, q.get)

CH1 = 'ch1'
CH2 = 'ch2'

class InputMultiplexerTestCase(unittest.TestCase):
    def test_multiple_input(self):
        imux = InputMultiplexer([CH1, CH2])

        # inactive on mux creation
        self.assertRaises(InactiveWritableError, imux[CH1].put, 'foo')
        self.assertRaises(InactiveWritableError, imux[CH2].put, 'foo')

        # activate
        imux[CH1].put(Begin)
        imux[CH1].put('foo')
        imux[CH1].put('foo-oo')

        # simple get.
        self.assertEqual(imux.get(), ('foo', CH1, ))

        imux[CH1].put(End)

        imux[CH2].put(Begin)
        imux[CH2].put('bar')
        imux[CH2].put('ba-ar')
        imux[CH2].put(End)

        self.assertEqual(imux.get(), ('foo-oo', CH1, ))
        self.assertEqual(imux.get(), ('bar', CH2, ))
        self.assertEqual(imux.get(), ('ba-ar', CH2, ))
        self.assertRaises(InactiveReadableError, imux.get)



if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from rdc.etl.hash import Hash
from rdc.etl.extra.simple import SimpleTransform


class TransformSimpleTestCase(unittest.TestCase):
    def _create_transform(self):
        return SimpleTransform()

    def test_remove(self):
        t = self._create_transform()
        t.remove('foo', 'boo')
        r = t.transform(Hash({'foo': 'bar', 'bar': 'baz', 'boo': 'hiya'}))
        self.assertIn('bar', r)
        self.assertNotIn('foo', r)
        self.assertNotIn('boo', r)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from collections import OrderedDict
from copy import copy

try:
    from thread import get_ident as _get_ident
except ImportError:
    from dummy_thread import get_ident as _get_ident

class Hash(OrderedDict):
    """
    Enhanced ordered dictionary type used to communicate named tuples between transform blocks.

    """

    def copy(self, datadict=None):
        o = copy(self)
        if datadict is not None:
            for k, v in datadict.items():
                o[k] = v
        return o

    def restrict(self, tester=None, renamer=None):
        """
        todo: simplify this as it does two things
        """
        keys = self.keys()
        for k in keys:
            if tester and not tester(k):
                del self[k]
            elif renamer:
                self[renamer(k)] = self[k]
                del self[k]
        return self

    def update(self, other=None, **kwargs):
        super(Hash, self).update(other, **kwargs)
        return self

    def remove(self, *keys):
        for key in keys:
            if key in self:
                del self[key]
        return self

    def get_values(self, keys):
        return [self[key] for key in keys]

    # BC
    def has(self, k, allow_none=False):
        return k in self and (allow_none or self[k] is not None)

    # BC
    def set(self, k, v):
        self[k] = v
        return self

    def __repr__(self, _repr_running={}):
        call_key = id(self), _get_ident()
        if call_key in _repr_running:
            return '...'
        _repr_running[call_key] = 1
        try:
            if not self:
                return 'H{}' % ()
            return 'H{%s}' % (', '.join(['%r: %r' % (k, v, ) for k, v in self.iteritems()]), )
        finally:
            del _repr_running[call_key]


    def checksum(self):
        from hashlib import sha1



# -*- coding: utf-8 -*-
#
# copyright 2012-2014 romain dorgueil
#
# licensed under the apache license, version 2.0 (the "license");
# you may not use this file except in compliance with the license.
# you may obtain a copy of the license at
#
#     http://www.apache.org/licenses/license-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the license is distributed on an "as is" basis,
# without warranties or conditions of any kind, either express or implied.
# see the license for the specific language governing permissions and
# limitations under the license.

import datetime
import time
import traceback
from threading import Thread
from rdc.etl import TICK
from rdc.etl.harness.base import BaseHarness
from rdc.etl.hash import Hash
from rdc.etl.io import InactiveReadableError, IO_TYPES, DEFAULT_INPUT_CHANNEL, DEFAULT_OUTPUT_CHANNEL, Begin, End
from rdc.etl.transform import Transform

class _IntSequenceGenerator(object):
    """Simple integer sequence generator."""

    def __init__(self):
        self.current = 0

    def get(self):
        return self.current

    def next(self):
        self.current += 1
        return self.current


class TransformThread(Thread):
    """Thread encapsulating a transformation. Handle errors."""

    __thread_counter = _IntSequenceGenerator()

    def __init__(self, transform, group=None, target=None, name=None, args=(), kwargs=None, verbose=None):
        super(TransformThread, self).__init__(group, target, name, args, kwargs, verbose)
        self.transform = transform
        self.__thread_number = self.__class__.__thread_counter.next()

    def handle_error(self, exc, tb):
        print str(exc) + '\n\n' + tb + '\n\n\n\n'

    @property
    def name(self):
        return self.transform.__name__ + '-' + str(self.__thread_number)

    def run(self):
        while True:
            try:
                self.transform.step()
            except KeyboardInterrupt as e:
                # todo send signal to harness, clean stop required. Needed ? read about threads and interrupt, only main may receive this.
                break
            except InactiveReadableError, e:
                # Terminated, exit loop.
                break
            except Exception, e:
                self.handle_error(e, traceback.format_exc())

        try:
            self.transform.step(finalize=True)
        except InactiveReadableError, e:
            # Obviously, ignore.
            pass
        except Exception, e:
            self.handle_error(e, traceback.format_exc())

    def __repr__(self):
        """Adds "alive" information to the transform representation."""
        return (self.is_alive() and '+' or '-') + ' ' + self.name + ' ' + self.transform.get_stats_as_string()


class ThreadedHarness(BaseHarness):
    """Builder for ETL job python callables, using threads for parallelization."""

    def __init__(self, debug=False, profile=False):
        super(ThreadedHarness, self).__init__()
        self.debug = debug
        self.profile = profile
        self.status = []
        self._transforms = {}
        self._threads = {}
        self._current_id = _IntSequenceGenerator()
        self._started_at = None
        self._finished_at = None

    def __call__(self):
        self._started_at = datetime.datetime.now()
        result = super(ThreadedHarness, self).__call__()
        self._finished_at = datetime.datetime.now()
        return result

    def get_threads(self):
        return self._threads.items()

    def get_transforms(self):
        return self._transforms.items()

    def add(self, transform):
        """Register a transformation, create a thread object to manage its future lifecycle."""
        id = self._current_id.next()
        self._transforms[id] = transform
        self._threads[id] = TransformThread(transform)
        return transform # BC, maybe id would be a better thing to return (todo 2.0, or even 1.0 before api freeze)

    def validate(self):
        """Validation of transform graph validity."""
        for id, transform in self._transforms.items():
            # Adds a special single empty hash queue to unplugged inputs
            for queue in transform._input.unplugged:
                queue.put(Begin)
                queue.put(Hash())
                queue.put(End)

            transform._output.put_all(Begin)

    def loop(self):
        """Starts all the thread and loop forever until they are all dead."""
        # todo healthcheck ? (cycles ... dead ends ... orphans ...)

        # start all threads
        for id, thread in self._threads.items():
            thread.start()

        # run initialization methods for statuses
        for status in self.status:
            status.initialize(self, debug=self.debug, profile=self.profile)

        # main loop until all threads are done
        while True:
            try:
                is_alive = False
                for id, thread in self._threads.items():
                    is_alive = is_alive or thread.is_alive()

                # communicate with the world
                for status in self.status:
                    status.update(self, debug=self.debug, profile=self.profile)

                # exit point
                if not is_alive:
                    break

                # take a nap. Time here determine how often status is updated, and the maximum waste of time after all
                # threads finished.
                time.sleep(TICK)
            except KeyboardInterrupt as e:
                # todo cleaner stop ?
                break

        # run finalization methods for statuses
        for status in self.status:
            status.finalize(self, debug=self.debug, profile=self.profile)

        # Wait for all transform threads to die
        for id, thread in self._threads.items():
            thread.join()


    def add_chain(self, *transforms, **kwargs):
        """Main helper method to add chains of transforms to this harness. You can plug the whole chain from and to
        other transforms by specifying `input` and `output` parameters.

        The transforms provided should not be bound yet.

        >>> h = ThreadedHarness()
        >>> t1, t2, t3 = Transform(), Transform(), Transform()
        >>> h.add_chain(t1, t2, t3)

        """
        if not len(transforms):
            raise Exception('At least one transform should be provided to form a chain.')

        input, output, input_channel, output_channel = None, None, None, None

        # Carefull! Input parameter should be an _output_ that we'll plug into our chain input.
        if 'input' in kwargs:
            input, input_channel = self.__find_output(kwargs['input'])

        # Carefull! Output parameter should be an _input_ that we'll plug into our chain input.
        if 'output' in kwargs:
            output, output_channel = self.__find_input(kwargs['output'])

        # Register the transformations and plug them together, as a chain.
        last_transform = None
        first_transform = transforms[0]
        for transform in transforms:
            if not transform.virgin:
                raise RuntimeError('You can\'t reuse a transform for now.')

            self.add(transform)

            if last_transform:
                self.__plug(last_transform._output, 0, transform._input, 0)

            last_transform = transform

        if input:
            # input contains the output of previous transform.
            self.__plug(input, input_channel, first_transform._input, 0)

        if output:
            # output contains the input we will plug our output into.
            self.__plug(last_transform._output, 0, output, output_channel)

    # Private stuff.

    def __find_input(self, mixed, default=DEFAULT_INPUT_CHANNEL):
        return self.__find_io_and_channel('input', mixed, default)

    def __find_output(self, mixed, default=DEFAULT_OUTPUT_CHANNEL):
        return self.__find_io_and_channel('output', mixed, default)

    def __find_io_and_channel(self, type, mixed, default=0):
        assert type in IO_TYPES, 'Invalid io type %r.' % (type, )

        if isinstance(mixed, Transform):
            io, channel = getattr(mixed, '_' + type), default
        elif isinstance(mixed, IO_TYPES[type]):
            io, channel = mixed, default
        elif len(mixed) == 1:
            io, channel = self.__find_io_and_channel(type, mixed[0], default)
        elif len(mixed) != 2:
            raise IOError('Unsupported %s given: %r.' % (type, mixed, ))
        else:
            io, channel = self.__find_io_and_channel(type, mixed[0], default=mixed[1])

        return io, channel

    def __plug(self, from_dmux, from_channel, to_mux, to_channel):
        to_mux.plug(from_dmux, channel=to_channel, dmux_channel=from_channel)





# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rdc.etl.io import STDIN
from rdc.etl.transform import Transform


class DatabaseCreateTable(Transform):
    table_name = None
    structure = ()
    drop_if_exists = False

    def __init__(self, engine, table_name=None, structure=None, drop_if_exists=None):
        super(DatabaseCreateTable, self).__init__()

        self.engine = engine
        self.table_name = table_name or self.table_name
        self.structure = structure or self.structure
        self.drop_if_exists = drop_if_exists or self.drop_if_exists
        self._executed = False

    def transform(self, hash, channel=STDIN):
        # this is a bit counterproductive, should tell that we don't change the flux, or delegate this to databaseload
        # or something
        if not self._executed:
            if self.drop_if_exists:
                query = 'DROP TABLE IF EXISTS %s;' % (self.table_name, )
                self.engine.execute(query)

            query = 'CREATE TABLE %s (%s)' % (
                self.table_name,
                ', \n'.join(['%s %s' % (n, t) for n, t in self.structure])
            )
            self.engine.execute(query)
            self._executed = True
        yield hash




# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from rdc.etl.extra.unittest import BaseTestCase
from rdc.etl.hash import Hash
from rdc.etl.transform.extract import Extract

INPUT_DATA = (
    {'foo': 'bar'},
    {'foo': 'baz'},
)

class TransformExtractTestCase(BaseTestCase):
    def test_base_class(self):
        t = Extract(extract=INPUT_DATA)
        self.assertStreamEqual(t(Hash()), INPUT_DATA)

    def test_base_class_generator(self):
        def generator():
            yield INPUT_DATA[0]
            yield INPUT_DATA[1]
        t = Extract(extract=generator)
        self.assertStreamEqual(t(Hash()), INPUT_DATA)

    def test_base_class_decorator(self):
        @Extract
        def my_iterable():
            return INPUT_DATA
        self.assertEqual(my_iterable.__name__, 'my_iterable')
        self.assertStreamEqual(my_iterable(Hash()), INPUT_DATA)

    def test_base_class_decorator_generator(self):
        @Extract
        def my_generator():
            yield INPUT_DATA[0]
            yield INPUT_DATA[1]
        self.assertEqual(my_generator.__name__, 'my_generator')
        self.assertStreamEqual(my_generator(Hash()), INPUT_DATA)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from rdc.etl.transform.flow import insert_sorted, default_comparator


class TransformFlowTestCase(unittest.TestCase):
    def test_insert_sorted(self):
        l = []

        # empty list
        insert_sorted(default_comparator, l, 'mmm', 'mmm')
        self.assertEqual(l, [('mmm', 'mmm', )])

        # append
        insert_sorted(default_comparator, l, 'nnn', 'nnn')
        self.assertEqual(l, [('mmm', 'mmm', ), ('nnn', 'nnn', )])

        # append
        insert_sorted(default_comparator, l, 'ooo', 'ooo')
        self.assertEqual(l, [('mmm', 'mmm', ), ('nnn', 'nnn', ), ('ooo', 'ooo', )])

        # before the last
        insert_sorted(default_comparator, l, 'nzz', 'nzz')
        self.assertEqual(l, [('mmm', 'mmm', ), ('nnn', 'nnn', ), ('nzz', 'nzz', ), ('ooo', 'ooo', )])

        # second
        insert_sorted(default_comparator, l, 'naa', 'naa')
        self.assertEqual(l, [('mmm', 'mmm', ), ('naa', 'naa', ), ('nnn', 'nnn', ), ('nzz', 'nzz', ), ('ooo', 'ooo', )])

        # first
        insert_sorted(default_comparator, l, 'aaa', 'aaa')
        self.assertEqual(l, [('aaa', 'aaa', ), ('mmm', 'mmm', ), ('naa', 'naa', ), ('nnn', 'nnn', ), ('nzz', 'nzz', ), ('ooo', 'ooo', )])


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#
# copyright 2012-2014 romain dorgueil
#
# licensed under the apache license, version 2.0 (the "license");
# you may not use this file except in compliance with the license.
# you may obtain a copy of the license at
#
#     http://www.apache.org/licenses/license-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the license is distributed on an "as is" basis,
# without warranties or conditions of any kind, either express or implied.
# see the license for the specific language governing permissions and
# limitations under the license.

from rdc.etl.error import AbstractError, ValidationError
from rdc.etl.transform import Transform


class SqlExec(Transform):
    def __init__(self, sql=None, db=None):
        # Use the callable name if provided
        if sql and not self._name:
            self._name = sql.__name__

        super(SqlExec, self).__init__()

        self.db = db
        self.sql = sql or self.sql


    def sql(self, hash, channel):
        raise AbstractError(self.sql)

    def validate(self):
        if self.db is None:
            raise ValidationError(self, '"db" positional argument is required.')

    def transform(self, hash, channel):
        try:
            params = self.sql(hash, channel)
            if params:
                sql, params = params[0], params[1:]
                self.db.execute(sql, params)
                hash['@{0}#{1}@status'.format(self.__name__, id(self))] = 'success'
        except Exception as e:
            hash['@{0}#{1}@status'.format(self.__name__, id(self))] = 'failure'
            raise
        finally:
            yield hash



# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from rdc.etl.hash import Hash


class HashTestCase(unittest.TestCase):
    def test_constructor_hash(self):
        hash = Hash({'foo': 'bar', 'bar': 'baz'})
        self.assertEqual(hash.get('foo'), 'bar')
        self.assertEqual(hash.get('bar'), 'baz')

    def test_constructor_zippedtuples(self):
        hash = Hash((('foo', 'bar', ), ('bar', 'baz', ), ))
        self.assertEqual(hash.get('foo'), 'bar')
        self.assertEqual(hash.get('bar'), 'baz')

    def test_constructor_default(self):
        hash = Hash()
        self.assertRaises(KeyError, hash.__getitem__, 'anything')

    def test_method_getter(self):
        hash = Hash({'foo': 'bar', 'bar': None})
        self.assertEqual(hash.get('foo', 'baz'), 'bar')
        self.assertEqual(hash.get('bar', 'baz'), None)
        self.assertEqual(hash.get('boo', 'foo'), 'foo')
        self.assertEqual(hash.get('boo'), None)

    def test_item_getter(self):
        hash = Hash({'foo': 'bar', 'bar': None})
        self.assertEqual(hash['foo'], 'bar')
        self.assertEqual(hash['bar'], None)
        self.assertRaises(KeyError, hash.__getitem__, 'boo')

    def test_method_setter(self):
        hash = Hash({'foo': 'bar'})
        hash.set('bar', 'heya')
        self.assertEqual(hash['bar'], 'heya')
        self.assertEqual(hash['foo'], 'bar')
        hash.set('foo', 'yoho')
        self.assertEqual(hash['foo'], 'yoho')
        self.assertEqual(hash.set('any', 'thing'), hash)

    def test_item_setter(self):
        hash = Hash({'foo': 'bar'})
        hash['bar'] = 'heya'
        self.assertEqual(hash['bar'], 'heya')
        self.assertEqual(hash['foo'], 'bar')
        hash['foo'] = 'yoho'
        self.assertEqual(hash['foo'], 'yoho')

    def test_method_in(self):
        hash = Hash({'foo': 'bar', 'bar': None})
        self.assertEqual(hash.has('foo'), True)
        self.assertEqual(hash.has('bar'), False)
        self.assertEqual(hash.has('bar', True), True)
        self.assertEqual(hash.has('baz'), False)

    def test_operator_in(self):
        hash = Hash({'foo': 'bar', 'bar': None})
        self.assertEqual('foo' in hash, True)
        self.assertEqual('bar' in hash, True)
        self.assertEqual('baz' in hash, False)

    def test_copy(self):
        h1 = Hash({'foo': 'bar', 'bar': 'baz', })
        h2 = h1.copy()
        h3 = h1.copy({'bar': 'oh my bar'})
        h1['foo'] = 'original foo'
        h2['foo'] = 'new foo'
        self.assertEqual(h1['foo'], 'original foo')
        self.assertEqual(h2['foo'], 'new foo')
        self.assertEqual(h3['foo'], 'bar')
        self.assertEqual(h3['bar'], 'oh my bar')

    def test_restrict(self):
        h = Hash({'foo': 'bar', 'bar': 'baz', })
        h.restrict(tester=lambda k: k in ('bar', ))
        self.assertTrue(not 'foo' in h)
        self.assertTrue('bar' in h)

    def test_restrict_with_renamer(self):
        h = Hash({'foo': 'bar', 'bar': 'baz', })
        h.restrict(tester=lambda k: k in ('bar', ), renamer=lambda k: k.upper())
        self.assertTrue(not 'foo' in h)
        self.assertTrue(not 'bar' in h)
        self.assertTrue('BAR' in h)
        self.assertEqual(h['BAR'], 'baz')

    def test_remove(self):
        h = Hash({'foo': 'bar', 'bar': 'baz', 'baz': 'boo', })
        h.remove('foo', 'baz')
        self.assertTrue(not 'foo' in h)
        self.assertTrue('bar' in h)
        self.assertTrue(not 'baz' in h)

    def test_get_values(self):
        h = Hash({'foo': 'bar', 'bar': 'baz', 'baz': 'boo', })
        self.assertEquals(h.get_values(('baz', 'foo', 'bar', )), ['boo', 'bar', 'baz', ])

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import OrderedDict

import unittest
from rdc.etl.extra.unittest import BaseTestCase
from rdc.etl.hash import Hash
from rdc.etl.transform.map import Map
from rdc.etl.transform.map.xml import XmlMap


class TransformMapXmlTestCase(BaseTestCase):
    def test_base_class_decorator(self):
        @XmlMap
        def xml_map(item):
            f = lambda path: item.findtext(path)
            return OrderedDict((
                ('id', item.attrib['id'], ),
                ('name', f('name'), ),
                ('value', ';'.join((i.text for i in item.findall('values/data'))), ),
            ))
        xml_map.xpath = './path/to/items/item'

        self.assertStreamEqual(
            xml_map((Hash({'_': '''<root>
                <path>
                    <to>
                        <items>
                            <item id="one">
                                <name>foo</name>
                                <values>
                                    <data>bar</data>
                                </values>
                            </item>
                            <item id="two">
                                <name>bar</name>
                                <values>
                                    <data>baz</data>
                                </values>
                            </item>
                            <item id="three">
                                <name>baz</name>
                                <values>
                                    <data>toto</data>
                                    <data>titi</data>
                                </values>
                            </item>
                        </items>
                    </to>
                </path>
            </root>'''}))), (
                OrderedDict((('id', 'one'), ('name', 'foo'), ('value', 'bar'), ), ),
                OrderedDict((('id', 'two'), ('name', 'bar'), ('value', 'baz'), ), ),
                OrderedDict((('id', 'three'), ('name', 'baz'), ('value', 'toto;titi'), ), ),
            ))


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rdc.etl.io import STDIN
from rdc.etl.transform import Transform
from rdc.etl.util import filter_html


def _apply_filter(value, hash, filter):
    # method filter
    if isinstance(filter, str):
        return getattr(value, filter)()

    # multi filter (gets the hash along with value
    elif hasattr(filter, '_is_multi') and filter._is_multi:
        return filter(value, hash)

    # simple standard filter, just a callable that transforms a value
    return filter(value)


class _SimpleItemTransformationDescriptor(object):
    def __init__(self, getter=None, *filters):
        self.getter = getter
        self.filters = list(filters)
        self.conditions = []

    def filter(self, filter):
        self.filters.append(filter)
        return self

    def filter_multi(self, filter):
        filter._is_multi = True
        self.filters.append(filter)
        return self

    def save_as(self, k, filter=None):
        """ Save the currently filtered value in another field, in its current state. You can still apply filters after
        this, but it's used to persist a partially transformed value.

        :param k: key to save under
        :return:
        """

        def _filter(v, h, k=k, filter=filter):
            h[k] = _apply_filter(v, h, filter) if filter is not None else v
            return v

        return self.filter_multi(_filter)

    def filter_html(self):
        self.filters.append(filter_html)
        return self

    def if_none(self, field=None):
        """
        TODO document this and add a generic if_

        Mark a field transformation to apply only the current value of a field is none. If no field is provided, then
        it looks at the target field.
        """

        def condition(hash, name, field=field):
            field = field or name
            return (not field in hash) or (hash[field] is None)

        self.conditions.insert(0, condition)
        return self

    def prepend(self, *fields, **options):
        cond = options.get('cond', None)
        postfix = options.get('postfix', None)
        separator = options.get('separator', '')

        # default conditions
        if cond is None:
            cond = lambda v: v and len(v)
        elif not callable(cond):
            cond = lambda v: cond

        def _filter(v, h, fields=fields, cond=cond, postfix=postfix, separator=separator):
            out = separator.join([h.get(field) for field in fields if cond(h.get(field))])
            if len(out) and postfix:
                out = out + postfix
            return out + (v or '')

        return self.filter_multi(_filter)

    def append(self, *fields, **options):
        cond = options.get('cond', None)
        prefix = options.get('prefix', None)
        separator = options.get('separator', '')

        # default conditions
        if cond is None:
            cond = lambda v: v and len(v)
        elif not callable(cond):
            cond = lambda v: cond

        def _filter(v, h, fields=fields, cond=cond, prefix=prefix, separator=separator):
            out = separator.join([h.get(field) for field in fields if cond(h.get(field))])
            if len(out) and prefix:
                out = prefix + out
            return (v or '') + out

        return self.filter_multi(_filter)

    def set_getter(self, getter):
        self.getter = getter
        return self

    def __call__(self, hash):
        if isinstance(self.getter, str):
            _name = self.getter

            def getter(o):
                try:
                    return o[_name]
                except KeyError, e:
                    return None

            getter.func_name = 'get_' + str(_name)

        elif isinstance(self.getter, unicode):
            _name = self.getter.encode('utf-8')

            def getter(o):
                try:
                    return o[_name]
                except KeyError, e:
                    return None

            getter.func_name = 'get_' + str(_name)

        else:
            getter = self.getter
            _name = repr(self.getter)

        value = getter(hash)

        for filter in self.filters:
            value = _apply_filter(value, hash, filter)

        return value


class SimpleTransform(Transform):
    """SimpleTransform is an attempt to make a trivial transformation easy to build, using fluid APIs and a lot of easy
    shortcuts to apply filters to some fields.

    The API is not stable and this will probably go into an "extra" module later.

    Example:

    >>> t = SimpleTransform()

    Apply "upper" method on "name" field, and store it back in "name" field.

    >>> t.add('name').filter('upper') # doctest: +ELLIPSIS
    <rdc.etl.extra.simple._SimpleItemTransformationDescriptor object at ...>

    Apply the lambda to "description" field content, and store it into the "full_description" field.

    >>> t.add('full_description', 'description').filter(lambda v: 'Description: ' + v) # doctest: +ELLIPSIS
    <rdc.etl.extra.simple._SimpleItemTransformationDescriptor object at ...>

    Remove the previously defined "useless" descriptor. This does not remove the "useless" fields into transformed
    hashes, it is only usefull to override some parent stuff.

    >>> t.useless = 'foo'
    >>> t.delete('useless')

    Mark the "notanymore" field for deletion upon transform. Output hashes will not anymore contain this field./

    >>> t.remove('notanymore')

    Add a field (output hashes will contain this field, all with the same "foo bar" value).

    >>> t.test_field = 'foo bar'

    .. todo::

        XXX the way it's implemented is very error prone, as we need to exclude "builtins" manually from __dict__
        iteration. A separate dict for transformed items would be better, as we don't use the "descriptor" ability of
        attribute classes anyway.

    """

    DescriptorClass = _SimpleItemTransformationDescriptor

    def __init__(self, *filters):
        super(SimpleTransform, self).__init__()
        self._filters = list(filters)

    def transform(self, hash, channel=STDIN):
        for name, value_getter in self.__dict__.items():
            if name[0] == '_' or name in ('INPUT_CHANNELS', 'OUTPUT_CHANNELS', 'transform' ):
                continue

            try:
                conditions = list(value_getter.conditions)
            except Exception, e:
                conditions = []

            can_update = True
            for condition in conditions:
                if not condition(hash, name):
                    can_update = False
                    break

            if can_update:
                if callable(value_getter):
                    hash[name] = value_getter(hash)
                else:
                    hash[name] = value_getter

        for filter in self._filters:
            hash = filter(hash)

        return hash

    def add(self, name, getter=None):
        if getter is None:
            getter = name
        descr = self.DescriptorClass(getter)
        setattr(self, name, descr)
        return descr

    def delete(self, name):
        """
        Unset a descriptor.
        """
        delattr(self, name)

    def remove(self, *names):
        """
        Removes a field in hash, using a post transform filter.
        """
        self.filter(lambda t, names=names: t.remove(*names))

    def filter(self, filter):
        """
        Adds a post transform filter.
        """
        self._filters.append(filter)
        return self




# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import types
from abc import ABCMeta, abstractmethod
from rdc.etl import H
from rdc.etl.error import AbstractError
from rdc.etl.hash import Hash
from rdc.etl.io import STDIN, STDOUT, STDERR, InputMultiplexer, OutputDemultiplexer, End
from rdc.etl.stat import Statisticable
from rdc.etl.util import Timer


class ITransform:
    __metaclass__ = ABCMeta

    @abstractmethod
    def transform(self, hash, channel=STDIN):
        """All input rows that comes to one of this transform's input channels will be passed to this method. If you
        only have one input channel, you can safely ignore the channel value, although you'll need it in method
        prototype."""
        raise AbstractError(self.transform)


class Transform(ITransform, Statisticable):
    """Base class and decorator for transformations.

    .. automethod:: transform

    .. attribute:: INPUT_CHANNELS

        List of input channel names.

    .. attribute:: OUTPUT_CHANNELS

        List of output channel names

    Example::

        >>> @Transform
        ... def my_transform(hash, channel=STDIN):
        ...     yield hash.copy({'foo': hash['foo'].upper()})

        >>> print list(my_transform(
        ...         H(('foo', 'bar'), ('bar', 'alpha')),
        ...         H(('foo', 'baz'), ('bar', 'omega')),
        ...     ))
        [H{'foo': 'BAR', 'bar': 'alpha'}, H{'foo': 'BAZ', 'bar': 'omega'}]

    """

    INPUT_CHANNELS = (STDIN, )
    OUTPUT_CHANNELS = (STDOUT, STDERR, )
    _name = None

    def __init__(self, transform=None, input_channels=None, output_channels=None):
        # Use the callable name if provided
        if transform and not self._name:
            self._name = transform.__name__

        self.INPUT_CHANNELS = input_channels or self.INPUT_CHANNELS
        self.OUTPUT_CHANNELS = output_channels or self.OUTPUT_CHANNELS

        self._input = InputMultiplexer(self.INPUT_CHANNELS)
        self._output = OutputDemultiplexer(self.OUTPUT_CHANNELS)
        self._exec_time = 0.0
        self._exec_count = 0

        self._booted = False
        self._initialized = False
        self._finalized = False

        self.transform = transform or self.transform

    def __call__(self, *stream, **options):
        channel = options['channel'] if 'channel' in options else STDIN

        for hash in stream:
            if not isinstance(hash, Hash):
                hash = Hash(hash)

            for line in self.transform(hash, channel):
                yield line

    # ITransform implementation

    def transform(self, hash, channel=STDIN):
        """Core transformation method that will be called for each input data row."""
        raise AbstractError(self.transform)

    # IO related

    def step(self, finalize=False):
        if not self._booted:
            # todo find something to make this work
            self.boot()
            self._booted = True

        if not self._initialized:
            self._initialized = True
            self.__execute_and_handle_output(self.initialize)

        try:
            # Pull data from the first available input channel (blocking)
            data, channel = self._input.get()
            # Execute actual transformation
            self.__execute_and_handle_output(self.transform, data, channel)
        finally:
            if finalize and not self._finalized:
                self._finalized = True
                self.__execute_and_handle_output(self.finalize)
                self._output.put_all(End)

    def boot(self):
        """Just before transformation is started, validate everything is ready."""
        pass

    def initialize(self):
        """If you need to execute code before any item is transformed, this is the place."""
        pass

    def finalize(self):
        """If you need to execute code after all items are transformed, this is the place. It's especially usefull for
        buffering transformations, or other blocking types."""
        pass

    @property
    def virgin(self):
        """Whether or not this transformation already contains a yucca (spéciale dédicace)."""
        return not self._initialized and not self._finalized

    # Name, statistics and representation logic. This is basic but important, as it will serve visualisation/debugging
    # purpose.

    @property
    def __name__(self):
        return self._name or type(self).__name__

    @__name__.setter
    def __name__(self, value):
        self._name = value

    def get_local_stats(self, debug=False, profile=False):
        if profile:
            return (
                (u'τ', '%.2fs' % (self._exec_time, ), ),
                (u'ε', self._exec_count, ),
                (u'τ.ε⁻¹', ((self._exec_count > 0) and (u'%.1fms' % (1000 * self._exec_time / self._exec_count, )) or u'∞'), ),
                (u'ε.τ⁻¹', ((self._exec_time > 0) and (u'%.1f/s' % (self._exec_count / self._exec_time, )) or u'∞'), ),
            )
        return ()

    def get_stats(self, debug=False, profile=False):
        stats = itertools.chain(
            self._input.stats,
            self._output.stats,
            self.get_local_stats(debug=debug, profile=profile)
        )
        return (
            (name, stat, ) for name, stat in stats
        )


    def __repr__(self):
        return u'<{0} {1}>'.format(self.__name__, self.get_unicode_stats())

    # Private
    def __execute_and_handle_output(self, callable, *args, **kwargs):
        """Runs a transformation callable with given args/kwargs and flush the result into the right
        output channel."""

        timer = Timer()
        with timer:
            results = callable(*args, **kwargs)
        self._exec_time += timer.duration

        # Put data onto output channels
        if isinstance(results, types.GeneratorType):
            while True:
                timer = Timer()
                with timer:
                    try:
                        result = results.next()
                    except StopIteration as e:
                        break
                self._exec_time += timer.duration
                self._exec_count += 1
                self._output.put(result)
        elif results is not None:
            self._exec_count += 1
            self._output.put(results)
        else:
            self._exec_count += 1





# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rdc.etl import DEFAULT_FIELD
from rdc.etl.transform.extract import Extract
from rdc.etl.util import create_http_reader, create_file_reader, cached_property


class FileExtract(Extract):
    """
    Extract data from a file into a field.

    .. attribute:: uri

        The path for source file. Can be either an absolute/relative filesystem path or an HTTP/HTTPS resource.

    .. attribute:: output_field

        The field that will contain file content. Use the topic (`_`) field by default.

    """
    uri = None
    output_field = DEFAULT_FIELD

    def __init__(self, uri=None, output_field=None):
        super(FileExtract, self).__init__()

        self.uri = uri or self.uri
        self.output_field = output_field or self.output_field

    @cached_property
    def reader(self):
        if self.uri is None:
            raise RuntimeError('No URI configured in ' + self.__class__.__name__ + ' transformation.')

        if self.uri.find('http://') == 0 or self.uri.find('https://') == 0:
            return create_http_reader(self.uri)
        else:
            return create_file_reader(self.uri)

    @cached_property
    def content(self):
        return self.reader()

    def extract(self):
        yield {
            self.output_field: self.content
        }




# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cgi
import HTMLParser
import time
import re
import requests
import types
import unidecode
from BeautifulSoup import BeautifulSoup
from blessings import Terminal as _Terminal
from datetime import datetime
from lxml import etree
from webapp2 import cached_property

VALID_HTML_TAGS = ['br']

# default unicode character mapping (you may not see some chars, leave as is )
char_map = {u'À': 'A', u'Á': 'A', u'Â': 'A', u'Ã': 'A', u'Ä': 'Ae', u'Å': 'A', u'Æ': 'A', u'Ā': 'A', u'Ą': 'A',
            u'Ă': 'A', u'Ç': 'C', u'Ć': 'C', u'Č': 'C', u'Ĉ': 'C', u'Ċ': 'C', u'Ď': 'D', u'Đ': 'D', u'È': 'E',
            u'É': 'E', u'Ê': 'E', u'Ë': 'E', u'Ē': 'E', u'Ę': 'E', u'Ě': 'E', u'Ĕ': 'E', u'Ė': 'E', u'Ĝ': 'G',
            u'Ğ': 'G', u'Ġ': 'G', u'Ģ': 'G', u'Ĥ': 'H', u'Ħ': 'H', u'Ì': 'I', u'Í': 'I', u'Î': 'I', u'Ï': 'I',
            u'Ī': 'I', u'Ĩ': 'I', u'Ĭ': 'I', u'Į': 'I', u'İ': 'I', u'Ĳ': 'IJ', u'Ĵ': 'J', u'Ķ': 'K', u'Ľ': 'K',
            u'Ĺ': 'K', u'Ļ': 'K', u'Ŀ': 'K', u'Ł': 'L', u'Ñ': 'N', u'Ń': 'N', u'Ň': 'N', u'Ņ': 'N', u'Ŋ': 'N',
            u'Ò': 'O', u'Ó': 'O', u'Ô': 'O', u'Õ': 'O', u'Ö': 'Oe', u'Ø': 'O', u'Ō': 'O', u'Ő': 'O', u'Ŏ': 'O',
            u'Œ': 'OE', u'Ŕ': 'R', u'Ř': 'R', u'Ŗ': 'R', u'Ś': 'S', u'Ş': 'S', u'Ŝ': 'S', u'Ș': 'S', u'Š': 'S',
            u'Ť': 'T', u'Ţ': 'T', u'Ŧ': 'T', u'Ț': 'T', u'Ù': 'U', u'Ú': 'U', u'Û': 'U', u'Ü': 'Ue', u'Ū': 'U',
            u'Ů': 'U', u'Ű': 'U', u'Ŭ': 'U', u'Ũ': 'U', u'Ų': 'U', u'Ŵ': 'W', u'Ŷ': 'Y', u'Ÿ': 'Y', u'Ý': 'Y',
            u'Ź': 'Z', u'Ż': 'Z', u'Ž': 'Z', u'à': 'a', u'á': 'a', u'â': 'a', u'ã': 'a', u'ä': 'ae', u'ā': 'a',
            u'ą': 'a', u'ă': 'a', u'å': 'a', u'æ': 'ae', u'ç': 'c', u'ć': 'c', u'č': 'c', u'ĉ': 'c', u'ċ': 'c',
            u'ď': 'd', u'đ': 'd', u'è': 'e', u'é': 'e', u'ê': 'e', u'ë': 'e', u'ē': 'e', u'ę': 'e', u'ě': 'e',
            u'ĕ': 'e', u'ė': 'e', u'ƒ': 'f', u'ĝ': 'g', u'ğ': 'g', u'ġ': 'g', u'ģ': 'g', u'ĥ': 'h', u'ħ': 'h',
            u'ì': 'i', u'í': 'i', u'î': 'i', u'ï': 'i', u'ī': 'i', u'ĩ': 'i', u'ĭ': 'i', u'į': 'i', u'ı': 'i',
            u'ĳ': 'ij', u'ĵ': 'j', u'ķ': 'k', u'ĸ': 'k', u'ł': 'l', u'ľ': 'l', u'ĺ': 'l', u'ļ': 'l', u'ŀ': 'l',
            u'ñ': 'n', u'ń': 'n', u'ň': 'n', u'ņ': 'n', u'ŉ': 'n', u'ŋ': 'n', u'ò': 'o', u'ó': 'o', u'ô': 'o',
            u'õ': 'o', u'ö': 'oe', u'ø': 'o', u'ō': 'o', u'ő': 'o', u'ŏ': 'o', u'œ': 'oe', u'ŕ': 'r', u'ř': 'r',
            u'ŗ': 'r', u'ś': 's', u'š': 's', u'ť': 't', u'ù': 'u', u'ú': 'u', u'û': 'u', u'ü': 'ue', u'ū': 'u',
            u'ů': 'u', u'ű': 'u', u'ŭ': 'u', u'ũ': 'u', u'ų': 'u', u'ŵ': 'w', u'ÿ': 'y', u'ý': 'y', u'ŷ': 'y',
            u'ż': 'z', u'ź': 'z', u'ž': 'z', u'ß': 'ss', u'ſ': 'ss', u'Α': 'A', u'Ά': 'A', u'Ἀ': 'A', u'Ἁ': 'A',
            u'Ἂ': 'A', u'Ἃ': 'A', u'Ἄ': 'A', u'Ἅ': 'A', u'Ἆ': 'A', u'Ἇ': 'A', u'ᾈ': 'A', u'ᾉ': 'A', u'ᾊ': 'A',
            u'ᾋ': 'A', u'ᾌ': 'A', u'ᾍ': 'A', u'ᾎ': 'A', u'ᾏ': 'A', u'Ᾰ': 'A', u'Ᾱ': 'A', u'Ὰ': 'A', u'Ά': 'A',
            u'ᾼ': 'A', u'Β': 'B', u'Γ': 'G', u'Δ': 'D', u'Ε': 'E', u'Έ': 'E', u'Ἐ': 'E', u'Ἑ': 'E', u'Ἒ': 'E',
            u'Ἓ': 'E', u'Ἔ': 'E', u'Ἕ': 'E', u'Έ': 'E', u'Ὲ': 'E', u'Ζ': 'Z', u'Η': 'I', u'Ή': 'I', u'Ἠ': 'I',
            u'Ἡ': 'I', u'Ἢ': 'I', u'Ἣ': 'I', u'Ἤ': 'I', u'Ἥ': 'I', u'Ἦ': 'I', u'Ἧ': 'I', u'ᾘ': 'I', u'ᾙ': 'I',
            u'ᾚ': 'I', u'ᾛ': 'I', u'ᾜ': 'I', u'ᾝ': 'I', u'ᾞ': 'I', u'ᾟ': 'I', u'Ὴ': 'I', u'Ή': 'I', u'ῌ': 'I',
            u'Θ': 'TH', u'Ι': 'I', u'Ί': 'I', u'Ϊ': 'I', u'Ἰ': 'I', u'Ἱ': 'I', u'Ἲ': 'I', u'Ἳ': 'I', u'Ἴ': 'I',
            u'Ἵ': 'I', u'Ἶ': 'I', u'Ἷ': 'I', u'Ῐ': 'I', u'Ῑ': 'I', u'Ὶ': 'I', u'Ί': 'I', u'Κ': 'K', u'Λ': 'L',
            u'Μ': 'M', u'Ν': 'N', u'Ξ': 'KS', u'Ο': 'O', u'Ό': 'O', u'Ὀ': 'O', u'Ὁ': 'O', u'Ὂ': 'O', u'Ὃ': 'O',
            u'Ὄ': 'O', u'Ὅ': 'O', u'Ὸ': 'O', u'Ό': 'O', u'Π': 'P', u'Ρ': 'R', u'Ῥ': 'R', u'Σ': 'S', u'Τ': 'T',
            u'Υ': 'Y', u'Ύ': 'Y', u'Ϋ': 'Y', u'Ὑ': 'Y', u'Ὓ': 'Y', u'Ὕ': 'Y', u'Ὗ': 'Y', u'Ῠ': 'Y', u'Ῡ': 'Y',
            u'Ὺ': 'Y', u'Ύ': 'Y', u'Φ': 'F', u'Χ': 'X', u'Ψ': 'PS', u'Ω': 'O', u'Ώ': 'O', u'Ὠ': 'O', u'Ὡ': 'O',
            u'Ὢ': 'O', u'Ὣ': 'O', u'Ὤ': 'O', u'Ὥ': 'O', u'Ὦ': 'O', u'Ὧ': 'O', u'ᾨ': 'O', u'ᾩ': 'O', u'ᾪ': 'O',
            u'ᾫ': 'O', u'ᾬ': 'O', u'ᾭ': 'O', u'ᾮ': 'O', u'ᾯ': 'O', u'Ὼ': 'O', u'Ώ': 'O', u'ῼ': 'O', u'α': 'a',
            u'ά': 'a', u'ἀ': 'a', u'ἁ': 'a', u'ἂ': 'a', u'ἃ': 'a', u'ἄ': 'a', u'ἅ': 'a', u'ἆ': 'a', u'ἇ': 'a',
            u'ᾀ': 'a', u'ᾁ': 'a', u'ᾂ': 'a', u'ᾃ': 'a', u'ᾄ': 'a', u'ᾅ': 'a', u'ᾆ': 'a', u'ᾇ': 'a', u'ὰ': 'a',
            u'ά': 'a', u'ᾰ': 'a', u'ᾱ': 'a', u'ᾲ': 'a', u'ᾳ': 'a', u'ᾴ': 'a', u'ᾶ': 'a', u'ᾷ': 'a', u'β': 'b',
            u'γ': 'g', u'δ': 'd', u'ε': 'e', u'έ': 'e', u'ἐ': 'e', u'ἑ': 'e', u'ἒ': 'e', u'ἓ': 'e', u'ἔ': 'e',
            u'ἕ': 'e', u'ὲ': 'e', u'έ': 'e', u'ζ': 'z', u'η': 'i', u'ή': 'i', u'ἠ': 'i', u'ἡ': 'i', u'ἢ': 'i',
            u'ἣ': 'i', u'ἤ': 'i', u'ἥ': 'i', u'ἦ': 'i', u'ἧ': 'i', u'ᾐ': 'i', u'ᾑ': 'i', u'ᾒ': 'i', u'ᾓ': 'i',
            u'ᾔ': 'i', u'ᾕ': 'i', u'ᾖ': 'i', u'ᾗ': 'i', u'ὴ': 'i', u'ή': 'i', u'ῂ': 'i', u'ῃ': 'i', u'ῄ': 'i',
            u'ῆ': 'i', u'ῇ': 'i', u'θ': 'th', u'ι': 'i', u'ί': 'i', u'ϊ': 'i', u'ΐ': 'i', u'ἰ': 'i', u'ἱ': 'i',
            u'ἲ': 'i', u'ἳ': 'i', u'ἴ': 'i', u'ἵ': 'i', u'ἶ': 'i', u'ἷ': 'i', u'ὶ': 'i', u'ί': 'i', u'ῐ': 'i',
            u'ῑ': 'i', u'ῒ': 'i', u'ΐ': 'i', u'ῖ': 'i', u'ῗ': 'i', u'κ': 'k', u'λ': 'l', u'μ': 'm', u'ν': 'n',
            u'ξ': 'ks', u'ο': 'o', u'ό': 'o', u'ὀ': 'o', u'ὁ': 'o', u'ὂ': 'o', u'ὃ': 'o', u'ὄ': 'o', u'ὅ': 'o',
            u'ὸ': 'o', u'ό': 'o', u'π': 'p', u'ρ': 'r', u'ῤ': 'r', u'ῥ': 'r', u'σ': 's', u'ς': 's', u'τ': 't',
            u'υ': 'y', u'ύ': 'y', u'ϋ': 'y', u'ΰ': 'y', u'ὐ': 'y', u'ὑ': 'y', u'ὒ': 'y', u'ὓ': 'y', u'ὔ': 'y',
            u'ὕ': 'y', u'ὖ': 'y', u'ὗ': 'y', u'ὺ': 'y', u'ύ': 'y', u'ῠ': 'y', u'ῡ': 'y', u'ῢ': 'y', u'ΰ': 'y',
            u'ῦ': 'y', u'ῧ': 'y', u'φ': 'f', u'χ': 'x', u'ψ': 'ps', u'ω': 'o', u'ώ': 'o', u'ὠ': 'o', u'ὡ': 'o',
            u'ὢ': 'o', u'ὣ': 'o', u'ὤ': 'o', u'ὥ': 'o', u'ὦ': 'o', u'ὧ': 'o', u'ᾠ': 'o', u'ᾡ': 'o', u'ᾢ': 'o',
            u'ᾣ': 'o', u'ᾤ': 'o', u'ᾥ': 'o', u'ᾦ': 'o', u'ᾧ': 'o', u'ὼ': 'o', u'ώ': 'o', u'ῲ': 'o', u'ῳ': 'o',
            u'ῴ': 'o', u'ῶ': 'o', u'ῷ': 'o', u'¨': '', u'΅': '', u'᾿': '', u'῾': '', u'῍': '', u'῝': '', u'῎': '',
            u'῞': '', u'῏': '', u'῟': '', u'῀': '', u'῁': '', u'΄': '', u'΅': '', u'`': '', u'῭': '', u'ͺ': '',
            u'᾽': '', u'А': 'A', u'Б': 'B', u'В': 'V', u'Г': 'G', u'Д': 'D', u'Е': 'E', u'Ё': 'E', u'Ж': 'ZH',
            u'З': 'Z', u'И': 'I', u'Й': 'I', u'К': 'K', u'Л': 'L', u'М': 'M', u'Н': 'N', u'О': 'O', u'П': 'P',
            u'Р': 'R', u'С': 'S', u'Т': 'T', u'У': 'U', u'Ф': 'F', u'Х': 'KH', u'Ц': 'TS', u'Ч': 'CH', u'Ш': 'SH',
            u'Щ': 'SHCH', u'Ы': 'Y', u'Э': 'E', u'Ю': 'YU', u'Я': 'YA', u'а': 'A', u'б': 'B', u'в': 'V', u'г': 'G',
            u'д': 'D', u'е': 'E', u'ё': 'E', u'ж': 'ZH', u'з': 'Z', u'и': 'I', u'й': 'I', u'к': 'K', u'л': 'L',
            u'м': 'M', u'н': 'N', u'о': 'O', u'п': 'P', u'р': 'R', u'с': 'S', u'т': 'T', u'у': 'U', u'ф': 'F',
            u'х': 'KH', u'ц': 'TS', u'ч': 'CH', u'ш': 'SH', u'щ': 'SHCH', u'ы': 'Y', u'э': 'E', u'ю': 'YU', u'я': 'YA',
            u'Ъ': '', u'ъ': '', u'Ь': '', u'ь': '', u'ð': 'd', u'Ð': 'D', u'þ': 'th', u'Þ': 'TH',
            u'ა': 'a', u'ბ': 'b', u'გ': 'g', u'დ': 'd', u'ე': 'e', u'ვ': 'v', u'ზ': 'z', u'თ': 't', u'ი': 'i',
            u'კ': 'k', u'ლ': 'l', u'მ': 'm', u'ნ': 'n', u'ო': 'o', u'პ': 'p', u'ჟ': 'zh', u'რ': 'r', u'ს': 's',
            u'ტ': 't', u'უ': 'u', u'ფ': 'p', u'ქ': 'k', u'ღ': 'gh', u'ყ': 'q', u'შ': 'sh', u'ჩ': 'ch', u'ც': 'ts',
            u'ძ': 'dz', u'წ': 'ts', u'ჭ': 'ch', u'ხ': 'kh', u'ჯ': 'j', u'ჰ': 'h'}


def slugify(s, strip=False):
    u"""
    Simple slug filter, that has no knowledge of diacritics. Prefer slughifi (see below) to this method for good slugs,
    even if for simple languages like english this may be enough (and probably faster).

    >>> text = u"C'est déjà l'été."
    >>> slugify(text)
    'c-est-deja-l-ete-'

    """
    str = re.sub(r'\W+', '-', unidecode.unidecode(s).lower())
    if strip:
        str = re.sub('(^-+|-+$)', '', str)
    return str


def replace_char(m):
    char = m.group()
    if char_map.has_key(char):
        return char_map[char]
    else:
        return char


def unaccent(value):
    """
    Replace diacritics with their ascii counterparts.
    """
    # unicodification
    if type(value) != types.UnicodeType:
        value = unicode(value, 'utf-8', 'ignore')

    # try to replace chars
    value = re.sub('[^a-zA-Z0-9\\s\\-]{1}', replace_char, value)

    return value.encode('ascii', 'ignore')


def slughifi(value, do_slugify=True, overwrite_char_map=None, strip=False):
    u"""
    High Fidelity slugify - slughifi.py, v 0.1

    This was found somewhere on internet, and slightly adapted for our needs.

    Examples :

    >>> text = u"C'est déjà l\'été."
    >>> slughifi(text)
    'c-est-deja-l-ete-'
    >>> slughifi(text, overwrite_char_map={"'": '-',})
    'c-est-deja-l-ete-'
    >>> slughifi(text, do_slugify=False)
    'C-est deja l-ete.'

    """

    # unicodification
    if type(value) != types.UnicodeType:
        value = unicode(value, 'utf-8', 'ignore')

    # overwrite chararcter mapping
    if overwrite_char_map:
        char_map.update(overwrite_char_map)

    # try to replace chars
    value = re.sub('[^a-zA-Z0-9\\s\\-]{1}', replace_char, value)

    # apply ascii slugify
    if do_slugify:
        value = slugify(value, strip=strip)

    return value.encode('ascii', 'ignore')


def filter_html(value):
    """
    Simple filter that removes all html found and replace HTML line breaks by a simple line feed character.
    """
    if value is None:
        return None
    soup = BeautifulSoup(value)
    for tag in soup.findAll(True):
        if tag.name not in VALID_HTML_TAGS:
            tag.hidden = True
    return soup.renderContents().replace('  ', ' ').replace('\n', '').replace('<br />', '\n')


class Timer(object):
    """
    Context manager used to time execution of stuff.
    """

    def __enter__(self):
        self.__start = time.time()

    def __exit__(self, type, value, traceback):
        # Error handling here
        self.__finish = time.time()

    @property
    def duration(self):
        return self.__finish - self.__start

    def __str__(self):
        return str(int(self.duration * 1000) / 1000.0) + 's'


def create_http_reader(url):
    """
    Simple reader for an HTTP resource.
    """

    def http_reader():
        return requests.get(url).content

    return http_reader


def create_file_reader(path):
    """
    Simple reader for a local filesystem resource.
    """

    def file_reader():
        with open(path, 'rU') as f:
            return f.read()

    return file_reader

def sfloat(mixed, default=None):
    """Safe float cast."""
    try:
        return float(mixed)
    except:
        return default

def sint(mixed, default=None):
    """Safe int cast."""
    try:
        return int(mixed)
    except:
        return default

def sbool(mixed, default=None):
    """Safe boolean cast."""
    try:
        return bool(mixed)
    except:
        return default

# Exports
terminal = _Terminal()
html_escape = cgi.escape
def html_unescape(txt):
    if not isinstance(txt, unicode):
        try:
            txt = txt.decode('raw_unicode_escape')
        except:
            print txt
    return HTMLParser.HTMLParser().unescape(txt)
now = datetime.now
cached_property = cached_property
etree = etree




# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from abc import ABCMeta, abstractmethod
from copy import copy
from Queue import Queue
import itertools
from rdc.etl.error import AbstractError, InactiveReadableError, InactiveWritableError
from rdc.etl.hash import Hash

# Input channels
from rdc.etl.stat import Statisticable

STDIN = 0
STDIN2 = 1
STDIN3 = 2
SELECT = 10

# Output channels
STDOUT = 0
STDERR = -1
STDOUT2 = 1
STDOUT3 = 2
INSERT = 10
UPDATE = 11

# Default channels
DEFAULT_INPUT_CHANNEL = STDIN
DEFAULT_OUTPUT_CHANNEL = STDOUT

# Types
INPUT_TYPE = 'input'
OUTPUT_TYPE = 'output'
OTHER_TYPE = 'other'

# Human friendly channel names
CHANNEL_NAMES = {
    INPUT_TYPE: {
        STDIN: 'in',
        STDIN2: 'in2',
        STDIN3: 'in3',
        SELECT: 'select'
    },
    OUTPUT_TYPE: {
        STDOUT: 'out',
        STDERR: 'err',
        STDOUT2: 'out2',
        STDOUT3: 'out3',
        INSERT: 'insert',
        UPDATE: 'update',
    },
}


class Token(object):
    """Factory for signal oriented queue messages."""

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return '<%s>' % (self.name, )

# Begin token raises a message queue runlevel.
Begin = Token('Begin')

# End token lowers a message queue runlevel.
End = Token('End')

# Default buffer size for queues.
BUFFER_SIZE = 8192


class IReadable:
    """Interface for things you can read from."""

    __metaclass__ = ABCMeta

    @abstractmethod
    def get(self, block=True, timeout=None):
        """Read. Block/timeout are there for Queue compat."""
        raise AbstractError(self.get)


class IWritable:
    """Interface for things you can write to."""

    __metaclass__ = ABCMeta

    @abstractmethod
    def put(self, data, block=True, timeout=None):
        """Write. Block/timeout are there for Queue compat."""
        raise AbstractError(self.put)


class InputMultiplexer(IReadable, Statisticable):
    def __init__(self, channels):
        self.queues = dict([(channel, Input()) for channel in channels])
        self._plugged = set()

        # statistic related
        self._stats = dict([(channel, 0) for channel in channels])
        self._special_stats = dict()

    def get_stats(self, debug=False, profile=False):
        stats = itertools.chain(self._stats.iteritems(), self._special_stats.iteritems())
        return ((CHANNEL_NAMES[INPUT_TYPE][channel], stat) for channel, stat in stats)

    def get(self, block=True, timeout=True):
        """Gets a (data, channel) tuple from the first queue ready for it.
        """
        # todo documentation says .enpty() value is not reliable ... XXX
        while self.alive:
            for id, queue in self.queues.items():
                if queue.alive and not queue.empty():
                    # xxx useage of block/timeout here is wrong
                    data = queue.get(block, timeout)

                    # increment stat counter
                    if not isinstance(data, Token):
                        self._stats[id] += 1

                    return data, id

            # todo xxx take block and timeout into account
            time.sleep(0.2)

        raise InactiveReadableError('InputMultiplexer is terminated.')

    def plug(self, dmux, channel=DEFAULT_INPUT_CHANNEL, dmux_channel=DEFAULT_OUTPUT_CHANNEL):
        dmux.plug_into(self.queues[channel], channel=dmux_channel)
        self._plugged.add(channel)

    def __getitem__(self, item):
        if not item in self.queues:
            raise KeyError('No such input channel %r.' % (item, ))
        return self.queues[item]

    @property
    def alive(self):
        for channel, queue in self.queues.items():
            if queue.alive:
                return True
        return False

    @property
    def plugged(self):
        return [queue for channel, queue in self.queues.items() if channel in self._plugged]

    @property
    def unplugged(self):
        return [queue for channel, queue in self.queues.items() if channel not in self._plugged]


class OutputDemultiplexer(IWritable, Statisticable):
    def __init__(self, channels):
        self.channels = dict([(channel, []) for channel in channels])

        # statistic related
        self._stats = dict([(channel, 0) for channel in channels])
        self._special_stats = dict()

    def get_stats(self, debug=False, profile=False):
        stats = itertools.chain(self._stats.iteritems(), self._special_stats.iteritems())
        return ((CHANNEL_NAMES[OUTPUT_TYPE][channel], stat) for channel, stat in stats)

    def put(self, data, block=True, timeout=None):
        data, channel = self.__demux(data)

        if not channel in self.channels:
            raise IOError('Unknown channel %r.' % (channel, ))

        # increment stat counter
        if not isinstance(data, Token):
            self._stats[channel] += 1

        for target in self.channels[channel]:
            target.put(isinstance(data, Token) and data or copy(data), block, timeout)

    def put_all(self, data, block=True, timeout=None):
        for channel in self.channels:
            self.put((data, channel, ), block, timeout)

    def plug_into(self, target, channel):
        if not channel in self.channels:
            raise IOError('Unknown channel %r.' % (channel, ))
        if target in self.channels[channel]:
            raise ValueError('Channel already have this target plugged for channel %r.' % (channel, ))
        self.channels[channel].append(target)


    def __getitem__(self, item):
        if not item in self.channels:
            raise KeyError('No such output channel %r.' % (item, ))
        return self.channels[item]

    def __demux(self, data):
        if isinstance(data, Hash):
            return data, DEFAULT_OUTPUT_CHANNEL
        if len(data) == 1:
            return data, DEFAULT_OUTPUT_CHANNEL
        if len(data) == 2:
            return data[0], data[1]
        raise ValueError('Unintelligible message.')


class Input(Queue, IReadable, IWritable):
    def __init__(self, maxsize=BUFFER_SIZE):
        Queue.__init__(self, maxsize)

        self._runlevel = 0
        self._writable_runlevel = 0

    def put(self, data, block=True, timeout=None):
        # Begin token is a metadata to raise the input runlevel.
        if data == Begin:
            self._runlevel += 1
            self._writable_runlevel += 1
            return

        # Check we are actually able to receive data.
        if self._writable_runlevel < 1:
            raise InactiveWritableError('Cannot put() on an inactive IWritable.')

        if data == End:
            self._writable_runlevel -= 1

        return Queue.put(self, data, block, timeout)

    def get(self, block=True, timeout=None):
        if not self.alive:
            raise InactiveReadableError('Cannot get() on an inactive IReadable.')

        data = Queue.get(self, block, timeout)

        if data == End:
            self._runlevel -= 1
            if not self.alive:
                raise InactiveReadableError('Cannot get() on an inactive IReadable (runlevel just reached 0).')
            return self.get(block, timeout)

        return data

    def empty(self):
        self.mutex.acquire()
        while self._qsize() and self.queue[0] == End:
            self._runlevel -= 1
            Queue._get(self)
        self.mutex.release()

        return Queue.empty(self)

    @property
    def alive(self):
        return self._runlevel > 0


IO_TYPES = {
    INPUT_TYPE: InputMultiplexer,
    OUTPUT_TYPE: OutputDemultiplexer,
}




# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Extracts are transformations that generate output lines from something that is not one of the input channel.
As it will yield all data for each input row, the input given is usually only one empty line.

"""

from rdc.etl.error import AbstractError
from rdc.etl.io import STDIN
from rdc.etl.transform import Transform

class Extract(Transform):
    """Base class for extract transforms.

    .. attribute:: extract

        Generator, iterable or iterable-typed callable that is used as the data source. Often used as a shortcut to
        make fast prototypes of ETL processes from a dictionary, before going further with real data sources.

        Each iterator value should be something Hash.copy() can take as an argument.


    Example using a dict::

        >>> from rdc.etl.transform.extract import Extract

        >>> data = ({'foo': 'bar'}, {'foo': 'baz'}, )
        >>> my_extract = Extract(extract=data)

        >>> list(my_extract({}))
        [H{'foo': 'bar'}, H{'foo': 'baz'}]


    Example using a callable::

        >>> from rdc.etl.transform.extract import Extract

        >>> @Extract
        ... def my_extract():
        ...     return (
        ...         {'bar': 'baz'},
        ...         {'bar': 'boo'},
        ...     )

        >>> list(my_extract({}))
        [H{'bar': 'baz'}, H{'bar': 'boo'}]


    Example using a generator::

        >>> from rdc.etl.transform.extract import Extract

        >>> @Extract
        ... def my_extract():
        ...     yield {'bar': 'baz'}
        ...     yield {'bar': 'boo'}

        >>> print list(my_extract({}))
        [H{'bar': 'baz'}, H{'bar': 'boo'}]

    .. note::

        Whenever you can, prefer the generator approach so you're not blocking anything while computing remaining
        elements.

    """

    extract = []

    def __init__(self, extract=None):
        super(Extract, self).__init__()

        self.extract = extract or self.extract

        if hasattr(extract, '__name__'):
            self.__name__ = self.extract.__name__

    def extract(self):
        raise AbstractError(self.extract)

    def transform(self, hash, channel=STDIN):
        extracted_data = self.extract() if callable(self.extract) else self.extract

        if extracted_data:
            for line in extracted_data:
                yield hash.copy(line)





# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rdc.etl.transform.extract import Extract


class DatabaseExtract(Extract):
    """
    Extract data from a database using some raw SQL and yield one output line per query result.

    .. attribute:: engine

        The sqlalchemy engine to use for extraction.

    .. attribute:: query

        The database query that will be used to extract data from database. Should not contain OFFSET/LIMIT, nor ";".

    .. attribute:: pack_size

        The number of records to retrieve at a time (will be used to add OFFSET/LIMIT clauses to SQL).

    """

    query = 'SELECT 1'
    pack_size = 100

    def __init__(self, engine, query=None, limit=None):
        super(DatabaseExtract, self).__init__()

        self.engine = engine
        self.query = query or self.query
        self.limit = limit

    def extract(self):
        self.query = self.query.strip()
        if self.query[-1] == ';':
            self.query = self.query[0:-1]

        offset = 0
        while not self.limit or offset * self.pack_size < self.limit:
            query = self.query + ' LIMIT ' + str(self.pack_size) + ' OFFSET ' + str(offset * self.pack_size) + ';'
            results = self.engine.execute(query, use_labels=True).fetchall()
            if not len(results):
                break

            for row in results:
                yield row

            offset += 1




# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
from rdc.etl.transform.extract import Extract
from rdc.etl.extra.simple import SimpleTransform
from rdc.etl.harness.threaded import ThreadedHarness
from rdc.etl.util import terminal as t

Harness = partial(ThreadedHarness, debug=True, profile=True)

def _get_value(id):
    return 'Value for %d' % (id, )

def build_producer(name, count=3, get_value=None, value_name='value'):
    return Extract([{'producer': name, 'id': count - i, value_name: (get_value or _get_value)(count - i)} for i in range(0, count)])

def build_simple_transform(f='upper'):
    t = SimpleTransform()
    t.add('producer').filter(f)
    return t

def run(harness):
    print
    retval = harness()
    print ' `-> {0}: {1}'.format(t.bold(t.white('Return')), retval)
    print

__all__ = [build_producer, build_simple_transform, run, Harness, ]




from setuptools import setup, find_packages

tolines = lambda c: filter(None, map(lambda s: s.strip(), c.split('\n')))

def read(filename, flt=None):
    with open(filename) as f:
        content = f.read().strip()
        return flt(content) if callable(flt) else content

def requirements_filter(c):
    install_requires = []
    for requirement in tolines(c):
        _pound_pos = requirement.find('#')
        if _pound_pos != -1:
            requirement = requirement[0:_pound_pos].strip()
        if len(requirement):
            install_requires.append(requirement)
    return install_requires

version = read('version.txt')

setup(
    name='rdc.etl',
    namespace_packages = ['rdc'],
    version=version,
    description="Extract Transform Load (ETL) toolkit for python",
    long_description=read('README.rst'),
    classifiers=read('classifiers.txt', tolines),
    keywords='ETL Data-Integration',
    author='Romain Dorgueil',
    author_email='romain@dorgueil.net',
    url='http://etl.rdc.li/',
    download_url='https://github.com/rdcli/rdc.etl/tarball/' + version,
    license='Apache License, Version 2.0',
    packages=find_packages(exclude=['ez_setup', 'example', 'test']),
    include_package_data=True,
    install_requires=read('requirements.txt', requirements_filter),
    entry_points="""
    [paste.paster_create_template]
    etl_project=rdc.etl.extra.tools.template:ETLProjectTemplate
    """,
)



# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rdc.etl.io import STDIN
from rdc.etl.transform import Transform
from rdc.etl.transform.flow import default_comparator, insert_sorted


class Sort(Transform):
    """
    Fast'n'dirty implementation of a software "sort" transform. As we need to stack up all the stream rows in our
    instance before we can start freeing them, this is most of the time a bad idea to use this. Always prefer to sort
    your rows the earliest possible, for example in a DatabaseExtract.

    This still can be an acceptable solution for proof of concepts or small streams of data.

    An example of why you would like to sort data is for streams that _requires_ data to be sorted, for example a double
    input joiner (not existing at the time of writing this).

    :attr:`key`
        A tuple of keys on which to sort data.

    """

    comparator = None

    def __init__(self, key, comparator=None):
        super(Sort, self).__init__()

        self.key = key
        self.comparator = comparator or default_comparator

    def initialize(self):
        self._sorted = []

    def transform(self, hash, channel=STDIN):
        key = hash.get_values(self.key)
        insert_sorted(self.comparator, self._sorted, key, hash)

    def finalize(self):
        for key, value in self._sorted:
            yield value




# -*- coding: utf-8 -*-
#
# copyright 2012-2014 romain dorgueil
#
# licensed under the apache license, version 2.0 (the "license");
# you may not use this file except in compliance with the license.
# you may obtain a copy of the license at
#
#     http://www.apache.org/licenses/license-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the license is distributed on an "as is" basis,
# without warranties or conditions of any kind, either express or implied.
# see the license for the specific language governing permissions and
# limitations under the license.

from abc import ABCMeta, abstractmethod
from rdc.etl.error import AbstractError

class IStatisticable:
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_stats(self, debug=False, profile=False):
        raise AbstractError(self.get_stats)

    @abstractmethod
    def get_unicode_stats(self, debug=False, profile=False):
        raise AbstractError(self.get_unicode_stats)


class Statisticable(IStatisticable):
    def get_unicode_stats(self, debug=False, profile=False):
        return u' '.join(
            (u'{0}={1}'.format(name, cnt) for name, cnt in self.get_stats(debug=debug, profile=profile) if cnt > 0)
        )

    # XXX BC remove this ASAP
    @property
    def stats(self):
        return self.get_stats()

    @property
    def stats_str(self):
        return self.get_unicode_stats()

    def get_stats_as_string(self):
        return self.get_unicode_stats()